"""Implements a pipeline for importing reference data from a raster image."""
from typing import List, Optional, Tuple

import fs
import numpy as np
from pydantic import Field, validator

from eolearn.core import CreateEOPatchTask, EONode, EOWorkflow, FeatureType, OverwritePermission, SaveTask
from eolearn.features.feature_manipulation import SpatialResizeTask
from eolearn.features.utils import ResizeLib, ResizeMethod, ResizeParam
from eolearn.io import ImportFromTiffTask

from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..utils.filter import get_patches_with_missing_features
from ..utils.types import Feature
from ..utils.validators import optional_field_validator, parse_dtype


class ResizeSchema(BaseSchema):
    """How to resize the tiff data after adding it to EOPatches."""

    parameters: Tuple[ResizeParam, Tuple[float, float]] = Field(
        description=(
            "Specify the resize parameters in the same way as for `SpatialResizeTask`. Examples:"
            ' `["resolution", [10, 20]]` for changing resolution from 10 to 20,'
            ' `["new_size", [500, 1000]]` for making the imported data of size (500, 1000).'
        )
    )
    method: ResizeMethod = ResizeMethod.LINEAR
    library: ResizeLib = ResizeLib.PIL

    @validator("parameters")
    def parameters_parser(cls, v: Tuple[ResizeParam, Tuple[float, float]]) -> Tuple[ResizeParam, Tuple[float, float]]:
        kind, params = v
        if kind is ResizeParam.NEW_SIZE:
            params = (round(params[0]), round(params[1]))
        return kind, params


class ImportTiffPipeline(Pipeline):
    class Schema(Pipeline.Schema):
        output_folder_key: str = Field(description="The storage manager key of the output folder.")
        tiff_folder_key: str = Field(description="The storage manager key of the folder containing the tiff to import.")
        input_filename: str = Field(description="Name of tiff file to import.")
        output_feature: Feature = Field(description="Feature containing the imported tiff information.")
        no_data_value: float = Field(
            np.nan, description="Value assigned to undefined pixels, e.g. outside of given input image."
        )
        dtype: Optional[np.dtype] = Field(description="Custom dtype for the imported feature.")
        _parse_dtype = optional_field_validator("dtype", parse_dtype, pre=True)
        use_vsi: bool = Field(
            True,
            description="Whether to use the VSI for reading. Enabled by default as a remote filesystem is assumed.",
        )
        resize: Optional[ResizeSchema] = Field(
            description="Settings for SpatialResizeTask applied at the end. When omitted resizing is not performed."
        )

    config: Schema

    def filter_patch_list(self, patch_list: List[str]) -> List[str]:
        """EOPatches are filtered according to existence of new features."""
        filtered_patch_list = get_patches_with_missing_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            [self.config.output_feature, FeatureType.BBOX],
        )
        return filtered_patch_list

    def build_workflow(self) -> EOWorkflow:
        create_eopatch_node = EONode(CreateEOPatchTask())

        file_path = fs.path.join(self.storage.get_folder(self.config.tiff_folder_key), self.config.input_filename)
        import_task = ImportFromTiffTask(
            self.config.output_feature,
            file_path,
            no_data_value=self.config.no_data_value,
            filesystem=self.storage.filesystem,
            image_dtype=self.config.dtype,
            use_vsi=self.config.use_vsi,
        )
        import_node = EONode(import_task, inputs=[create_eopatch_node])

        resize_node = None
        if self.config.resize:
            resize_task = SpatialResizeTask(
                resize_parameters=self.config.resize.parameters,
                resize_method=self.config.resize.method,
                resize_library=self.config.resize.library,
            )
            resize_node = EONode(resize_task, inputs=[import_node])

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            compress_level=1,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            config=self.sh_config,
            features=[self.config.output_feature, FeatureType.BBOX],
        )
        save_node = EONode(save_task, inputs=[resize_node or import_node])

        return EOWorkflow.from_endnodes(save_node)



"""
Area managers working with UTM CRS
"""
import logging
from typing import Any, List, Tuple, cast

from geopandas import GeoDataFrame
from pydantic import Field

from sentinelhub import Geometry, UtmZoneSplitter

from ...utils.general import reduce_to_coprime
from ...utils.grid import GridTransformation, create_transformations
from .base import AreaManager

LOGGER = logging.getLogger(__name__)


class UtmZoneAreaManager(AreaManager):
    """Area manager that splits grid per UTM zones"""

    class Schema(AreaManager.Schema):
        patch_size_x: int = Field(description="A width of each EOPatch in meters")
        patch_size_y: int = Field(description="A height of each EOPatch in meters")
        patch_buffer_x: float = Field(
            0, description="Number of meters by which to increase the tile size to left and right."
        )
        patch_buffer_y: float = Field(
            0, description="Number of meters by which to increase the tile size to up and down."
        )
        offset_x: float = Field(0, description="An offset of tiling grid in horizontal dimension")
        offset_y: float = Field(0, description="An offset of tiling grid in vertical dimension")

    config: Schema

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.absolute_buffer = self.config.patch_buffer_x, self.config.patch_buffer_y

    def _get_grid_filename_params(self) -> List[object]:
        """A list of parameters specific to UTM zone splitting"""
        return [
            self.config.patch_size_x,
            self.config.patch_size_y,
            self.config.patch_buffer_x,
            self.config.patch_buffer_y,
            self.config.offset_x,
            self.config.offset_y,
        ]

    def _create_new_split(self, area_geometry: Geometry) -> List[GeoDataFrame]:
        """Uses UtmZoneSplitter to create a grid"""
        LOGGER.info("Splitting area geometry into UTM zone grid")
        splitter = UtmZoneSplitter(
            [area_geometry.geometry],
            crs=area_geometry.crs,
            bbox_size=(self.config.patch_size_x, self.config.patch_size_y),
            offset=(self.config.offset_x, self.config.offset_y),
        )

        bbox_list = splitter.get_bbox_list()
        if self.absolute_buffer != (0, 0):
            bbox_list = [bbox.buffer(self.absolute_buffer, relative=False) for bbox in bbox_list]

        info_list = splitter.get_info_list()
        for tile_info in info_list:
            tile_info["index_n"] = tile_info["index"]
            tile_info["total_num"] = len(bbox_list)

        return self._to_dataframe_grid(bbox_list, info_list, ["index_n", "index_x", "index_y", "total_num"])

    def transform_grid(self, target_area_manager: AreaManager) -> List[GridTransformation]:
        """Calculates how grid should be transformed into another area manager."""
        self._check_target_compatibility(target_area_manager)
        target_area_manager = cast(UtmZoneAreaManager, target_area_manager)

        source_grid = self.get_grid(add_bbox_column=True)
        target_grid = target_area_manager.get_grid(add_bbox_column=True)

        source_size = self.config.patch_size_x, self.config.patch_size_y
        target_size = target_area_manager.config.patch_size_x, target_area_manager.config.patch_size_y

        sizes_x, sizes_y = zip(source_size, target_size)
        target_shape, source_shape = zip(reduce_to_coprime(*sizes_x), reduce_to_coprime(*sizes_y))

        source_grid = self._add_patch_positions(source_grid, self.config, source_shape)
        target_grid = self._add_patch_positions(target_grid, target_area_manager.config, target_shape)

        return create_transformations(source_grid, target_grid, match_columns=["position_x", "position_y"])

    @staticmethod
    def _add_patch_positions(grid: List[GeoDataFrame], config: Schema, shape: Tuple[int, int]) -> List[GeoDataFrame]:
        """Adds columns of values that define how sub-tiles should be spatially divided into groups. Members of each
        group will have the same values in the newly added columns and will spatially form a regular subgrid of given
        shape.

        The process works with the lower left vertex of the bounding box. First it reverts buffer and offset shift.
        This way each vertex will have coordinates that are an integer multiple of patch size because this is the way
        `UtmZoneSplitter` created the grid. In the next step the vertex values are divided by patch size and the given
        shape of the subgroup. This produces exactly the desired indices that define groups.
        """
        translate_x = config.patch_buffer_x - config.offset_x
        translate_y = config.patch_buffer_y - config.offset_y
        scale_x = config.patch_size_x * shape[0]
        scale_y = config.patch_size_y * shape[1]

        for gdf in grid:
            gdf["position_x"] = gdf.BBOX.apply(lambda bbox: round(bbox.min_x + translate_x) // scale_x)
            gdf["position_y"] = gdf.BBOX.apply(lambda bbox: round(bbox.min_y + translate_y) // scale_y)

        return grid

    def _check_target_compatibility(self, target_area_manager: AreaManager) -> None:
        """Checks if source and target managers are compatible to define a transformation between them."""
        if not isinstance(target_area_manager, UtmZoneAreaManager):
            raise NotImplementedError(
                f"Grid transformation is only supported into another {self.__class__.__name__}, but got "
                f"{target_area_manager.__class__.__name__}"
            )

        for param in ["offset_x", "offset_y"]:
            if getattr(self.config, param) != getattr(target_area_manager.config, param):
                raise NotImplementedError(
                    "Cannot transform grid because source and target area managers have different values of config "
                    f"parameter {param}."
                )



"""
Download pipeline that works with Sentinel Hub batch service
"""
import logging
from typing import Any, DefaultDict, Dict, List, Literal, Optional, Tuple

from pydantic import Field

from sentinelhub import (
    BatchRequest,
    BatchRequestStatus,
    BatchTileStatus,
    BatchUserAction,
    DataCollection,
    MimeType,
    MosaickingOrder,
    ResamplingType,
    SentinelHubBatch,
    SentinelHubRequest,
    monitor_batch_analysis,
    monitor_batch_job,
    read_data,
)

from ..core.area.batch import BatchAreaManager
from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..utils.types import Path, TimePeriod
from ..utils.validators import field_validator, optional_field_validator, parse_data_collection, parse_time_period

LOGGER = logging.getLogger(__name__)


class InputDataSchema(BaseSchema):
    """Parameter structure for a single data collection used in a batch request."""

    data_collection: DataCollection = Field(
        description=(
            "Data collection from which data will be downloaded. See `utils.validators.parse_data_collection` for more"
            " info on input options."
        )
    )
    _validate_data_collection = field_validator("data_collection", parse_data_collection, pre=True)

    time_period: Optional[TimePeriod]
    _validate_time_period = optional_field_validator("time_period", parse_time_period, pre=True)

    resampling_type: ResamplingType = Field(
        ResamplingType.NEAREST, description="A type of downsampling and upsampling used by Sentinel Hub service"
    )
    maxcc: Optional[float] = Field(ge=0, le=1, description="Maximal cloud coverage filter.")
    mosaicking_order: Optional[MosaickingOrder] = Field(description="The mosaicking order used by Sentinel Hub service")
    other_params: dict = Field(
        default_factory=dict,
        description=(
            "Additional parameters to be passed to SentinelHubRequest.input_data method as other_args parameter."
        ),
    )


class BatchDownloadPipeline(Pipeline):
    """Pipeline to start and monitor a Sentinel Hub batch job"""

    class Schema(Pipeline.Schema):
        area: BatchAreaManager.Schema

        output_folder_key: str = Field(
            description="Storage manager key pointing to the path where batch results will be saved."
        )

        inputs: List[InputDataSchema]
        evalscript_path: Path

        tiff_outputs: List[str] = Field(default_factory=list, description="Names of TIFF outputs of a batch job")
        save_userdata: bool = Field(
            False, description="A flag indicating if userdata.json should also be one of the results of the batch job."
        )
        batch_output_kwargs: dict = Field(
            default_factory=dict,
            description=(
                "Any other arguments to be added to a dictionary of parameters. Passed as `**kwargs` to the `output`"
                " method of `SentinelHubBatch` during the creation process."
            ),
        )

        analysis_only: bool = Field(
            False,
            description=(
                "If set to True it will only create a batch request and wait for analysis phase to finish. It "
                "will not start the actual batch job."
            ),
        )
        monitoring_sleep_time: int = Field(
            120,
            ge=60,
            description=(
                "How many seconds to sleep between two consecutive queries about status of tiles in a batch "
                "job. It should be at least 60 seconds."
            ),
        )
        monitoring_analysis_sleep_time: int = Field(
            10,
            ge=5,
            description=(
                "How many seconds to sleep between two consecutive queries about a status of a batch job analysis "
                "phase. It should be at least 5 seconds."
            ),
        )

        batch_id: str = Field(
            "",
            description=(
                "An ID of a batch job for this pipeline. If it is given the pipeline will just monitor the "
                "existing batch job. If it is not given it will create a new batch job."
            ),
        )
        patch_list: None = None
        input_patch_file: None = None
        skip_existing: Literal[False] = False

    config: Schema
    area_manager: BatchAreaManager

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.batch_client = SentinelHubBatch(config=self.sh_config)

        if self.area_manager.subsplit != (1, 1):
            raise ValueError(
                f"Cannot run {self.__class__.__name__} if {self.area_manager.__class__.__name__} is using a subsplit."
                " Set subsplit parameters to 1."
            )

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Procedure that uses Sentinel Hub batch service to download data to an S3 bucket."""
        batch_request = self._create_or_collect_batch_request()
        user_action = self._trigger_user_action(batch_request)

        if user_action is BatchUserAction.ANALYSE or (
            user_action is BatchUserAction.START and batch_request.status is not BatchRequestStatus.ANALYSIS_DONE
        ):
            LOGGER.info("Waiting to finish analyzing job with ID %s", batch_request.request_id)
            monitor_batch_analysis(
                batch_request,
                config=self.sh_config,
                sleep_time=self.config.monitoring_analysis_sleep_time,
            )

        self.ensure_batch_grid(batch_request.request_id)

        if self.config.analysis_only:
            return [], []

        LOGGER.info("Monitoring batch job with ID %s", batch_request.request_id)
        results = monitor_batch_job(
            batch_request,
            config=self.sh_config,
            sleep_time=self.config.monitoring_sleep_time,
            analysis_sleep_time=self.config.monitoring_analysis_sleep_time,
        )

        processed = self._get_tile_names_from_results(results, BatchTileStatus.PROCESSED)
        failed = self._get_tile_names_from_results(results, BatchTileStatus.FAILED)
        log_msg = f"Successfully downloaded {len(processed)} tiles"
        log_msg += f", but {len(failed)} tiles failed." if failed else "."
        LOGGER.info(log_msg)
        return processed, failed

    def _create_or_collect_batch_request(self) -> BatchRequest:
        """Either creates a new batch request or collects information about an existing one."""
        if not self.config.batch_id:
            batch_request = self._create_new_batch_request()
            LOGGER.info("Created a new batch request with ID %s", batch_request.request_id)
            return batch_request

        batch_request = self.batch_client.get_request(self.config.batch_id)
        batch_request.raise_for_status(status=[BatchRequestStatus.FAILED, BatchRequestStatus.CANCELED])
        LOGGER.info("Collected existing batch request with ID %s", batch_request.request_id)
        return batch_request

    def _create_new_batch_request(self) -> BatchRequest:
        """Defines a new batch request."""
        geometry = self.area_manager.get_area_geometry()

        responses = [
            SentinelHubRequest.output_response(tiff_output, MimeType.TIFF) for tiff_output in self.config.tiff_outputs
        ]
        if self.config.save_userdata:
            responses.append(SentinelHubRequest.output_response("userdata", MimeType.JSON))

        sentinelhub_request = SentinelHubRequest(
            evalscript=read_data(self.config.evalscript_path, data_format=MimeType.TXT),
            input_data=[
                SentinelHubRequest.input_data(
                    data_collection=input_config.data_collection,
                    time_interval=input_config.time_period,
                    upsampling=input_config.resampling_type,
                    downsampling=input_config.resampling_type,
                    maxcc=input_config.maxcc,
                    mosaicking_order=input_config.mosaicking_order,
                    other_args=input_config.other_params,
                )
                for input_config in self.config.inputs
            ],
            responses=responses,
            geometry=geometry,
        )

        data_folder = self.storage.get_folder(self.config.output_folder_key, full_path=True).rstrip("/")
        if not self.storage.is_on_aws():
            raise ValueError(f"The data folder path should be on s3 bucket, got {data_folder}")

        return self.batch_client.create(
            sentinelhub_request,
            tiling_grid=SentinelHubBatch.tiling_grid(
                grid_id=self.config.area.tiling_grid_id,
                resolution=self.config.area.resolution,
                buffer=(self.config.area.tile_buffer_x, self.config.area.tile_buffer_y),
            ),
            output=SentinelHubBatch.output(
                default_tile_path=f"{data_folder}/<tileName>/<outputId>.<format>", **self.config.batch_output_kwargs
            ),
            description=f"eo-grow - {self.__class__.__name__} pipeline with ID {self.pipeline_id}",
        )

    def _trigger_user_action(self, batch_request: BatchRequest) -> BatchUserAction:
        """According to status and configuration parameters decide what kind of user action to perform."""
        if self.config.analysis_only:
            if batch_request.status is BatchRequestStatus.CREATED:
                self.batch_client.start_analysis(batch_request)
                LOGGER.info("Triggered batch job analysis.")
                return BatchUserAction.ANALYSE

            status = None if batch_request.status is None else batch_request.status.value
            LOGGER.info("Didn't trigger analysis because current batch request status is %s.", status)
            return BatchUserAction.NONE

        if batch_request.status in [
            BatchRequestStatus.CREATED,
            BatchRequestStatus.ANALYSING,
            BatchRequestStatus.ANALYSIS_DONE,
        ]:
            self.batch_client.start_job(batch_request)
            LOGGER.info("Started running batch job.")
            return BatchUserAction.START

        if batch_request.status is BatchRequestStatus.PARTIAL:
            self.batch_client.restart_job(batch_request)
            LOGGER.info("Restarted partially failed batch job.")
            return BatchUserAction.START

        status = None if batch_request.status is None else batch_request.status.value
        LOGGER.info("Didn't trigger batch job because current batch request status is %s", status)
        return BatchUserAction.NONE

    def ensure_batch_grid(self, request_id: str) -> None:
        """This method ensures that area manager caches batch grid into the storage."""
        if self.area_manager.batch_id and self.area_manager.batch_id != request_id:
            raise ValueError(
                f"{self.area_manager.__class__.__name__} is set to use batch request with ID "
                f"{self.area_manager.batch_id} but {self.__class__.__name__} is using batch request with ID "
                f"{request_id}. Make sure that you use the same IDs."
            )
        self.area_manager.batch_id = request_id

        self.area_manager.cache_grid()

    @staticmethod
    def _get_tile_names_from_results(
        results: DefaultDict[BatchTileStatus, List[Dict]], tile_status: BatchTileStatus
    ) -> List[str]:
        """Collects tile names from a dictionary of batch tile results ordered by status"""
        tile_list = results[tile_status]
        return [tile["name"] for tile in tile_list]



"""
Modules with Ray-related utilities
"""
import logging

import ray

from .types import BoolOrAuto

LOGGER = logging.getLogger(__name__)


def handle_ray_connection(use_ray: BoolOrAuto = "auto") -> bool:
    """According to the given parameter it will try to connect to an existing Ray cluster.

    :param use_ray: Either a boolean flag or `"auto"` to define if the connection should be established or not.
    :return: `True` if connection is established and `False` otherwise.
    """
    if use_ray == "auto":
        try:
            _try_connect_to_ray()
            return True
        except ConnectionError:
            LOGGER.info("No Ray cluster found, will not use Ray.")
            return False

    if use_ray:
        _try_connect_to_ray()
        return True
    return False


def _try_connect_to_ray() -> None:
    """Try connecting and log if successful."""
    ray.init(address="auto", ignore_reinit_error=True)
    LOGGER.info("Connected to an existing Ray cluster.")



"""
Module with utilities for creating maps
"""
import logging
import os
import shutil
import subprocess
from tempfile import NamedTemporaryFile
from typing import Iterable, Literal, Optional

LOGGER = logging.getLogger(__name__)

GDAL_DTYPE_SETTINGS = {
    "uint8": "Byte",
    "uint16": "UInt16",
    "int16": "Int16",
    "float32": "Float32",
}
CogifyResamplingOptions = Literal[None, "NEAREST", "MODE", "AVERAGE", "BILINEAR", "CUBIC", "CUBICSPLINE", "LANCZOS"]
WarpResamplingOptions = Literal[
    None,
    "near",
    "bilinear",
    "cubic",
    "cubicspline",
    "lanczos",
    "average",
    "rms",
    "mode",
    "max",
    "min",
    "med",
    "q1",
    "q3",
    "sum",
]


def cogify_inplace(
    tiff_file: str,
    blocksize: int = 2048,
    nodata: Optional[float] = None,
    dtype: Literal[None, "int8", "int16", "uint8", "uint16", "float32"] = None,
    resampling: CogifyResamplingOptions = None,
    quiet: bool = True,
) -> None:
    """Make the (geotiff) file a cog
    :param tiff_file: .tiff file to cogify
    :param blocksize: block size of tiled COG
    :param nodata: value to be treated as nodata, default value is None
    :param dtype: output type of the in the resulting tiff, default is None
    :param resampling: The resampling method used to produce overviews. The defaults (when using None) are CUBIC for
        floats and NEAREST for integers.
    :param quiet: The process does not produce logs.
    """
    temp_file = NamedTemporaryFile()
    temp_file.close()

    cogify(
        tiff_file,
        temp_file.name,
        blocksize,
        nodata=nodata,
        dtype=dtype,
        overwrite=True,
        resampling=resampling,
        quiet=quiet,
    )
    # Note: by moving the file we also remove the one at temp_file.name
    shutil.move(temp_file.name, tiff_file)


def cogify(
    input_file: str,
    output_file: str,
    blocksize: int = 1024,
    nodata: Optional[float] = None,
    dtype: Literal[None, "int8", "int16", "uint8", "uint16", "float32"] = None,
    overwrite: bool = True,
    resampling: CogifyResamplingOptions = None,
    quiet: bool = True,
) -> None:
    """Create a cloud optimized version of input file

    :param input_file: File to cogify
    :param output_file: Resulting cog file
    :param blocksize: block size of tiled COG
    :param nodata: value to be treated as nodata, default value is None
    :param dtype: output type of the in the resulting tiff, default is None
    :param overwrite: If True overwrite the output file if it exists.
    :param resampling: The resampling method used to produce overviews. The defaults (when using None) are CUBIC for
        floats and NEAREST for integers.
    :param quiet: The process does not produce logs.
    """
    if input_file == output_file:
        raise OSError("Input file is the same as output file.")

    if os.path.exists(output_file):
        if overwrite:
            os.remove(output_file)
        else:
            raise OSError(f"{output_file} exists!")

    version = subprocess.check_output(("gdalinfo", "--version"), text=True).split(",")[0].replace("GDAL ", "")
    if version < "3.1.0":
        raise RuntimeError(
            f"The cogification process is configured for GDAL 3.1.0 and higher, but version {version} was detected.",
            RuntimeWarning,
        )

    gdaltranslate_options = (
        f"-of COG -co COMPRESS=DEFLATE -co BLOCKSIZE={blocksize} -co OVERVIEWS=IGNORE_EXISTING -co PREDICTOR=YES"
    )

    if resampling:
        gdaltranslate_options += f" -co RESAMPLING={resampling}"

    if quiet:
        gdaltranslate_options += " -q"

    if nodata is not None:
        gdaltranslate_options += f" -a_nodata {nodata}"

    if dtype is not None:
        gdaltranslate_options += f" -ot {GDAL_DTYPE_SETTINGS[dtype]}"

    subprocess.check_call(f"gdal_translate {gdaltranslate_options} {input_file} {output_file}", shell=True)


def merge_tiffs(
    input_filenames: Iterable[str],
    merged_filename: str,
    *,
    overwrite: bool = True,
    nodata: Optional[float] = None,
    dtype: Literal[None, "int8", "int16", "uint8", "uint16", "float32"] = None,
    warp_resampling: WarpResamplingOptions = None,
    quiet: bool = True,
) -> None:
    """Performs gdal_merge on a set of given geotiff images

    :param input_filenames: A sequence of input tiff image filenames
    :param merged_filename: Filename of merged tiff image
    :param overwrite: If True overwrite the output (merged) file if it exists
    :param delete_input: If True input images will be deleted at the end
    :param warp_resampling: The resampling method used when warping, useful for pixel misalignment. Defaults to NEAREST.
    :param quiet: The process does not produce logs.
    """
    gdalwarp_options = "-co BIGTIFF=YES -co compress=LZW -co TILED=YES"

    if overwrite:
        gdalwarp_options += " -overwrite"

    if quiet:
        gdalwarp_options += " -q"

    if warp_resampling:
        gdalwarp_options += f" -r {warp_resampling}"

    if nodata is not None:
        gdalwarp_options += f' -dstnodata "{nodata}"'

    if dtype is not None:
        gdalwarp_options += f" -ot {GDAL_DTYPE_SETTINGS[dtype]}"

    command = f"gdalwarp {gdalwarp_options} {' '.join(input_filenames)} {merged_filename}"
    subprocess.check_call(command, shell=True)


def extract_bands(
    input_file: str,
    output_file: str,
    bands: Iterable[int],
    overwrite: bool = True,
    compress: bool = False,
    quiet: bool = True,
) -> None:
    """Extract bands from given input file

    :param input_file: File containing all bands
    :param output_file: Resulting file with extracted bands
    :param bands: Sequence of bands to extract. Indexation starts at 0.
    :param overwrite: If True overwrite the output file if it exists.
    :param quiet: The process does not produce logs.
    """
    if not bands:
        raise ValueError("No bands were specified for extraction, undefined behaviour.")

    if input_file == output_file:
        raise OSError("Input file is the same as output file.")

    if os.path.exists(output_file):
        if overwrite:
            os.remove(output_file)
        else:
            raise OSError(f"{output_file} already exists. Set `overwrite` to true if it should be overwritten.")

    # gdal_translate starts indexation at 1
    translate_opts = " ".join(f" -b {band + 1}" for band in bands)
    if quiet:
        translate_opts += " -q"
    if compress:
        translate_opts += " -co compress=LZW"

    command = f"gdal_translate {translate_opts} {input_file} {output_file}"
    subprocess.check_call(command, shell=True)



"""
Module implementing export_maps pipelines
"""
import datetime as dt
import itertools as it
import logging
from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from typing import Dict, Iterable, List, Literal, Optional, Tuple

import fs
import fs.copy
import numpy as np
from fs.base import FS
from fs.tempfs import TempFS
from pydantic import Field
from tqdm.auto import tqdm

from eolearn.core import EONode, EOPatch, EOTask, EOWorkflow, FeatureType, LoadTask, linearly_connect_tasks
from eolearn.core.utils.fs import get_full_path, pickle_fs, unpickle_fs
from eolearn.core.utils.parallelize import parallelize
from eolearn.features import LinearFunctionTask
from eolearn.io import ExportToTiffTask
from sentinelhub import CRS, MimeType

from eogrow.core.config import RawConfig

from ..core.pipeline import Pipeline
from ..utils.map import CogifyResamplingOptions, WarpResamplingOptions, cogify_inplace, extract_bands, merge_tiffs
from ..utils.types import Feature

LOGGER = logging.getLogger(__name__)

TIMESTAMP_FORMAT = "%Y-%m-%dT%H-%M-%S"


class ExportMapsPipeline(Pipeline):
    """Pipeline to export a feature into a tiff map"""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(
            description="The storage manager key pointing to the input folder for the export maps pipeline."
        )
        output_folder_key: str = Field(
            description=(
                "The storage manager key pointing to the output folder for the maps in the export maps pipeline."
            )
        )

        feature: Feature
        map_name: Optional[str] = Field(regex=r".+\." + MimeType.TIFF.extension + r"?\b")  # noqa
        map_dtype: Literal["int8", "int16", "uint8", "uint16", "float32"]
        no_data_value: Optional[float] = Field(description="No data value to be passed to GeoTIFFs")
        scale_factor: Optional[float] = Field(description="Feature will be multiplied by this value at export")
        band_indices: Optional[List[int]] = Field(
            description="A list of band indices to be exported for the export feature. Default is all bands"
        )
        warp_resampling: WarpResamplingOptions = Field(
            None, description="The resampling method used when warping, useful for pixel misalignment"
        )

        cogify: bool = Field(
            False, description="Whether exported GeoTIFFs will be converted into Cloud Optimized GeoTIFFs (COG)"
        )
        cogification_resampling: CogifyResamplingOptions = Field(
            None, description="Which resampling to use in the cogification process for creating overviews."
        )

        force_local_copies: bool = Field(
            False,
            description=(
                "By default copying to local temporary folder will happen only if an AWS S3 path is configured."
                " With this parameter you force to always make copies."
            ),
        )
        interim_results_suffix: str = Field(
            "",
            description=(
                "Has no effect on end results. Adds a suffix to names of temporary files in order to avoid clashes"
                " from pipelines working in parallel on same maps (e.g. exporting same map for different timestamps)."
            ),
        )
        split_per_timestamp: bool = Field(
            True,
            description=(
                "Temporal features are by default exported as multiple TIFF files in a per-timestamp manner. Disabling "
                "this parameter results in a single TIFF with the same band order as the one of `ExportToTiffTask`."
            ),
        )
        skip_existing: Literal[False] = False
        merge_workers: Optional[int] = Field(
            description=(
                "How many workers are used to parallelize merging of TIFFs. Uses all cores (of head node) by default."
                "Decreasing this should help with memory and disk-space issues."
            )
        )

    config: Schema
    MERGED_MAP_NAME = "merged"

    def __init__(self, config: Schema, raw_config: Optional[RawConfig] = None):
        super().__init__(config, raw_config)

        self.map_name = self.config.map_name or f"{self.config.feature[1]}.{MimeType.TIFF.extension}"
        self.get_tiff_name = partial(get_tiff_name, self.map_name, suffix=self.config.interim_results_suffix)

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Extracts and merges the data from EOPatches into a TIFF file.

        1. Extracts data from EOPatches via workflow into per-EOPatch tiffs.
        2. For each UTM zone:
            - Prepares tiffs for merging (transfers to local if needed).
            - Performs temporal split of tiffs if needed (assumption that all eopatches share the same timestamps)
            - Merges the tiffs
            - Cogification is done if requested.
            - The output files are finalized (renamed/transferred) and per-EOPatch tiffs are cleaned.

        """

        successful, failed = super().run_procedure()

        if not successful:
            raise ValueError("Failed to extract tiff files from any of EOPatches.")

        feature_type, _ = self.config.feature
        output_folder = self.storage.get_folder(self.config.output_folder_key)
        crs_eopatch_dict = self.eopatch_manager.split_by_utm(successful)

        for crs, eopatch_list in crs_eopatch_dict.items():
            LOGGER.info("Processing UTM %d", crs.epsg)

            exported_tiff_paths = [
                fs.path.join(output_folder, self.get_tiff_name(patch_name)) for patch_name in eopatch_list
            ]
            filesystem, geotiff_paths = self._prepare_files(exported_tiff_paths)

            crs_output_folder = fs.path.join(output_folder, f"UTM_{crs.epsg}")
            filesystem.makedirs(crs_output_folder, recreate=True)

            if feature_type.is_timeless() or not self.config.split_per_timestamp:
                merged_map_name = self.get_tiff_name(self.MERGED_MAP_NAME, crs)
                combine_tiffs_jobs = [CombineTiffsJob(geotiff_paths, merged_map_name, time=None)]
            else:
                time_to_tiffs_map = self._split_patches_temporally(
                    filesystem, crs_output_folder, geotiff_paths, some_eopatch=eopatch_list[0], crs=crs
                )
                map_name_maker = partial(self.get_tiff_name, self.MERGED_MAP_NAME, crs)

                combine_tiffs_jobs = [
                    CombineTiffsJob(tiffs, map_name_maker(time), time) for time, tiffs in time_to_tiffs_map.items()
                ]

            LOGGER.info("Merging TIFF files.")
            parallelize(
                partial(self._combine_geotiffs, self.config, pickle_fs(filesystem)),
                combine_tiffs_jobs,
                workers=self.config.merge_workers,
            )

            LOGGER.info("Finalizing output.")
            merged_maps = {job.time: job.output_path for job in combine_tiffs_jobs}
            self._finalize_output(filesystem, merged_maps, crs_output_folder, exported_tiff_paths)

        return successful, failed

    def build_workflow(self) -> EOWorkflow:
        load_task = LoadTask(
            self.storage.get_folder(self.config.input_folder_key),
            filesystem=self.storage.filesystem,
            features=[self.config.feature, FeatureType.BBOX],
        )
        task_list: List[EOTask] = [load_task]

        if self.config.scale_factor is not None:
            rescale_task = LinearFunctionTask(self.config.feature, slope=self.config.scale_factor)
            task_list.append(rescale_task)

        export_to_tiff_task = ExportToTiffTask(
            self.config.feature,
            folder=self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            no_data_value=self.config.no_data_value,
            image_dtype=np.dtype(self.config.map_dtype),
            band_indices=self.config.band_indices,
        )
        task_list.append(export_to_tiff_task)

        return EOWorkflow(linearly_connect_tasks(*task_list))

    def get_execution_arguments(self, workflow: EOWorkflow) -> List[Dict[EONode, Dict[str, object]]]:
        exec_args = super().get_execution_arguments(workflow)
        nodes = workflow.get_nodes()
        for node in nodes:
            if isinstance(node.task, ExportToTiffTask):
                for patch_name, single_exec_dict in zip(self.patch_list, exec_args):
                    single_exec_dict[node] = dict(filename=self.get_tiff_name(patch_name))

        return exec_args

    def _prepare_files(self, geotiff_paths: List[str]) -> Tuple[FS, List[str]]:
        """Returns a filesystem and appropriate paths to the geotiffs to use in the pipeline.

        If required files are copied locally and a temporary filesystem object is returned.
        """
        make_local_copies = self.storage.is_on_aws() or self.config.force_local_copies

        if not make_local_copies:
            return self.storage.filesystem, geotiff_paths

        temp_fs = TempFS(identifier="_merge_maps_temp")
        temp_geotiff_paths = [fs.path.basename(path) for path in geotiff_paths]

        LOGGER.info("Copying tiffs to a temporary local folder %s", temp_fs.getsyspath("/"))
        parallelize(
            fs.copy.copy_file,
            it.repeat(self.storage.filesystem),
            geotiff_paths,
            it.repeat(temp_fs),
            temp_geotiff_paths,
            workers=None,
            multiprocess=False,
        )
        return temp_fs, temp_geotiff_paths

    def _split_patches_temporally(
        self, filesystem: FS, output_folder: str, geotiff_paths: List[str], some_eopatch: str, crs: CRS
    ) -> Dict[Optional[dt.datetime], List[str]]:
        """Splits multi-temporal tiffs into separate tiffs according to the temporal dimensions.

        The jobs for splitting tiffs are grouped together per input file. This ensures that two processes don't try
        to use GDAL on the same input file (avoiding IO blocks).

        Because the rest of the pipeline groups parallelization jobs per time, the function returns `time_to_tiffs_map`
        that groups output tiffs of the same timestamp.
        """
        LOGGER.info("Splitting TIFF files temporally.")
        num_bands, timestamps = self._extract_num_bands_and_timestamps(some_eopatch)  # assume eopatches share these

        def get_extraction_path(input_path: str, time: dt.datetime) -> str:
            """Ensures that after extraction the files have a time-suffix."""
            tiff_name = _strip_tiff_extension(fs.path.basename(input_path))
            return fs.path.join(output_folder, self.get_tiff_name(tiff_name, crs, time))

        time_to_tiffs_map: Dict[Optional[dt.datetime], List[str]] = defaultdict(list)
        grouped_split_jobs = []  # these are grouped per input file to avoid IO blocking of processes
        for tiff_path in geotiff_paths:
            tiff_sys_path = filesystem.getsyspath(tiff_path)

            jobs = []
            for i, time in enumerate(timestamps):
                extraction_path = get_extraction_path(tiff_path, time)
                bands = range(i * num_bands, (i + 1) * num_bands)
                jobs.append(SplitTiffsJob(tiff_sys_path, bands, filesystem.getsyspath(extraction_path)))

                time_to_tiffs_map[time].append(extraction_path)

            grouped_split_jobs.append(jobs)

        parallelize(self._execute_split_jobs, grouped_split_jobs, workers=None)
        parallelize(filesystem.remove, geotiff_paths, workers=None, multiprocess=False)

        return time_to_tiffs_map

    def _extract_num_bands_and_timestamps(self, eopatch_name: str) -> Tuple[int, List[dt.datetime]]:
        """Loads an eopatch to get information about number of bands and the timestamps."""
        path = fs.path.join(self.storage.get_folder(self.config.input_folder_key), eopatch_name)
        patch = EOPatch.load(path, (FeatureType.TIMESTAMP, self.config.feature), filesystem=self.storage.filesystem)
        if self.config.band_indices is not None:
            return len(self.config.band_indices), patch.timestamp
        return patch[self.config.feature].shape[-1], patch.timestamp

    @staticmethod
    def _execute_split_jobs(jobs: List["SplitTiffsJob"]) -> None:
        """Executes all the jobs for a specific tiff. This prevents parallel processes fighting over IO to a tiff."""
        for job in jobs:
            extract_bands(job.input_path, job.output_path, job.bands)

    @staticmethod
    def _combine_geotiffs(config: Schema, pickled_filesystem: bytes, job: "CombineTiffsJob") -> None:
        """Merges tiffs and cogifies them if needed. Also removes the pre-merge tiffs."""
        filesystem = unpickle_fs(pickled_filesystem)
        merge_tiffs(
            map(filesystem.getsyspath, job.input_paths),
            filesystem.getsyspath(job.output_path),
            nodata=config.no_data_value,
            dtype=config.map_dtype,
            warp_resampling=config.warp_resampling,
        )
        for tiff_path in job.input_paths:
            filesystem.remove(tiff_path)

        if config.cogify:
            cogify_inplace(
                filesystem.getsyspath(job.output_path),
                blocksize=1024,
                nodata=config.no_data_value,
                dtype=config.map_dtype,
                resampling=config.cogification_resampling,
            )

    def _finalize_output(
        self,
        filesystem: FS,
        merged_maps_paths: Dict[Optional[dt.datetime], str],
        output_folder: str,
        exported_tiff_paths: List[str],
    ) -> None:
        """Renames (or transfers in case of temporal FS) the files to the expected output files.

        If a temporal filesystem was used, the exported tiffs are cleaned from the storage and the filesystem closed.
        """
        self.storage.filesystem.makedirs(output_folder, recreate=True)
        for timestamp, map_path in tqdm(merged_maps_paths.items()):
            if timestamp is None:
                output_path = fs.path.join(output_folder, self.map_name)
            else:
                time_output_folder = fs.path.join(output_folder, timestamp.strftime(TIMESTAMP_FORMAT))
                self.storage.filesystem.makedirs(time_output_folder, recreate=True)
                output_path = fs.path.join(time_output_folder, self.map_name)

            fs.copy.copy_file(filesystem, map_path, self.storage.filesystem, output_path)
            filesystem.remove(map_path)

        LOGGER.info("Merged maps are saved in %s", get_full_path(self.storage.filesystem, output_folder))

        if isinstance(filesystem, TempFS):
            filesystem.close()

            # for non-tempfs runs this is already done right after merging
            LOGGER.info("Remove per-eopatch tiffs from storage.")
            parallelize(self.storage.filesystem.remove, exported_tiff_paths, workers=None, multiprocess=False)


@dataclass
class SplitTiffsJob:
    """Describes which bands of input path to extract to the output path."""

    input_path: str
    bands: Iterable[int]
    output_path: str


@dataclass
class CombineTiffsJob:
    """Describes which tiffs to merge, what the output path is, and the time the merged tiff represents.

    The time is relevant in order to correctly place the finalized tiff. If left out then the tiffs are either
    timeless or not split temporally.
    """

    input_paths: Iterable[str]
    output_path: str
    time: Optional[dt.datetime]


def _strip_tiff_extension(path: str) -> str:
    return path.replace(f".{MimeType.TIFF.extension}", "")


def get_tiff_name(
    map_name: str, name: str, crs: Optional[CRS] = None, time: Optional[dt.datetime] = None, suffix: str = ""
) -> str:
    """Creates a name of a geotiff image"""
    base = f"{_strip_tiff_extension(map_name)}_{name}"
    if crs is not None:
        base += f"_UTM_{crs.epsg}"
    if time is not None:
        base += f"_{time.strftime(TIMESTAMP_FORMAT)}"
    if suffix:
        base += f"_{suffix}"
    return f"{base}.{MimeType.TIFF.extension}"



"""
Module implementing management of configurations
"""
import base64
import copy
import json
import os
import re
from functools import reduce
from typing import Any, Callable, Dict, List, NewType, Optional, Set, cast

import fs.path
import rapidjson

from eolearn.core.utils.fs import get_base_filesystem_and_path, get_full_path, join_path

from ..utils.general import jsonify
from ..utils.meta import get_os_import_path

CrudeConfig = NewType("CrudeConfig", dict)
RawConfig = NewType("RawConfig", dict)


def encode_config_list(configs: List[CrudeConfig]) -> str:
    """Dumps a list of configs into a json and the encodes it with base64

    :return: A base64-encoded string
    """
    json_string = json.dumps(configs, default=jsonify)
    return base64.b64encode(json_string.encode()).decode()


def decode_config_list(encoded_config_list: str) -> List[CrudeConfig]:
    """Provides a list of config objects by decoding a base64-encoded string."""
    decoded_string = base64.b64decode(encoded_config_list.encode()).decode()
    decoded_list = json.loads(decoded_string)
    if not isinstance(decoded_list, list):
        raise ValueError(f"Encoding does not represent a list of configs. Got {type(decoded_list)}.")
    return decoded_list


def collect_configs_from_path(path: str, used_config_paths: Optional[Set[str]] = None) -> List[CrudeConfig]:
    """Loads and builds a list of config dictionaries defined by the parameters stored in files

    This function performs the 1st stage of language interpretation as described in
    eo-grow/documentation/config-language.md`.

    :param path: A full path where a config file is located
    :param used_config_paths: A helper parameter to prevent cyclic config imports.
    :return: A list of stage 1 dictionaries from which to build config object.
    """
    filesystem, path = get_base_filesystem_and_path(path)
    with filesystem.open(path, "r") as file_handle:
        config = rapidjson.load(file_handle, parse_mode=rapidjson.PM_COMMENTS)

    used_config_paths = copy.copy(used_config_paths) or set()
    used_config_paths.add(get_full_path(filesystem, path))

    full_folder_path = get_full_path(filesystem, fs.path.dirname(path))
    config = _recursive_apply_to_strings(config, lambda value: _resolve_config_paths(value, full_folder_path))

    config = _recursive_config_build(config, used_config_paths)

    if isinstance(config, dict):
        config = [config]
    if isinstance(config, list):
        return config
    raise ValueError(f"When interpreting config from {path} a dictionary or list was expected, got {type(config)}.")


def _recursive_config_build(config: object, used_config_paths: Set[str]) -> object:
    """Recursively builds a configuration object by replacing dictionary items in form of

        `'**key': 'file path to another config'`

    with a content of a configuration file, referenced with the path. The items from both configuration objects are
    joined recursively.
    """
    if isinstance(config, dict):
        joint_config = {}
        imported_configs: List[CrudeConfig] = []

        for key, value in config.items():
            if not isinstance(key, str):
                raise ValueError(f"Dictionary keys should always be strings, but found: {key}")

            if key.startswith("**"):
                if value in used_config_paths:
                    raise ValueError("Detected a cyclic import of configs")

                imported_config_list = collect_configs_from_path(value, used_config_paths=used_config_paths)
                if len(imported_config_list) != 1:
                    raise ValueError(
                        "Config lists cannot be imported inside configs. Found a config list when resolving key"
                        f" {key} for path {value}"
                    )
                imported_configs.append(imported_config_list[0])
            else:
                joint_config[key] = _recursive_config_build(value, used_config_paths)

        return reduce(recursive_config_join, imported_configs, joint_config)

    if isinstance(config, list):
        return [_recursive_config_build(value, used_config_paths) for value in config]

    return config


def interpret_config_from_dict(config: CrudeConfig, external_variables: Optional[Dict[str, Any]] = None) -> RawConfig:
    """Applies config language rules to a loaded config

    This function performs the 2nd stage of language interpretation as described in
    `eo-grow/documentation/config-language.md`.
    """
    _recursive_check_config(config)
    config_with_imports = _recursive_apply_to_strings(config, _resolve_import_paths)
    if not isinstance(config_with_imports, dict):
        raise ValueError(f"Interpretation resulted in object of type {type(config)} but a dictionary was expected.")
    config = cast(CrudeConfig, config_with_imports)

    variable_mapping = config.pop("variables", {})
    if external_variables:
        variable_mapping.update(external_variables)

    for variable_name in variable_mapping:
        if not re.fullmatch(r"\w+", variable_name):
            raise ValueError(f"Variable name {variable_name} contains illegal characters")

    config_with_variables = _recursive_apply_to_strings(
        config, lambda config_str: _resolve_variables(config_str, variable_mapping)
    )
    if not isinstance(config_with_variables, dict):
        raise ValueError(
            f"Interpretation resulted in object of type {type(config_with_variables)} but a dictionary was expected."
        )

    return cast(RawConfig, config_with_variables)


def interpret_config_from_path(path: str) -> RawConfig:
    """Loads from path in applies both steps of the config language."""
    configs = collect_configs_from_path(path)
    if len(configs) != 1:
        raise ValueError(f"The JSON file {path} was expected to contain a single dictionary, got {len(configs)}")
    return interpret_config_from_dict(configs[0])


def _resolve_config_paths(config_str: str, config_path: str) -> str:
    """Replaces `${config_path}` with an actual path"""
    new_config_str = re.sub(r"\${config_path}", config_path, config_str)

    if new_config_str != config_str:
        return join_path(new_config_str)
    return new_config_str


def _resolve_import_paths(config_str: str) -> str:
    """Replaces `${import_path:package.module}` with an actual local path"""
    new_config_str = re.sub(r"\${import_path:([\w.]+)}", _sub_import_path, config_str)

    if new_config_str != config_str:
        return join_path(new_config_str)
    return new_config_str


def _sub_import_path(match: re.Match) -> str:
    """Substitutes a regex match with a local file path to a module or a package"""
    import_module = match.group(1)
    return os.path.dirname(get_os_import_path(import_module))


def _resolve_variables(config_str: str, variable_mapping: Dict[str, str]) -> str:
    """Replaces `${var:variable_name}` with a replacement value defined under variables"""
    return re.sub(r"\${var:(\w+)}", lambda match: _sub_variable(match, variable_mapping), config_str)


def _sub_variable(match: re.Match, variable_mapping: Dict[str, str]) -> str:
    """Substitutes a regex match with a new variable according to the mapping"""
    variable_name = match.group(1)
    if variable_name in variable_mapping:
        return str(variable_mapping[variable_name])
    raise ValueError(f"Variable name '{variable_name}' doesn't exist in a config dictionary of variables.")


def _recursive_apply_to_strings(config: object, function: Callable) -> object:
    """Recursively applies a function on all string values (and not keys) of a nested config object"""
    if isinstance(config, dict):
        return {key: _recursive_apply_to_strings(value, function) for key, value in config.items()}

    if isinstance(config, list):
        return [_recursive_apply_to_strings(value, function) for value in config]

    if isinstance(config, str):
        return function(config)
    return config


def _recursive_check_config(config: object) -> None:
    """Recursively checks if the config satisfies basic conditions for being JSON serializable.

    :raises: ValueError
    """
    if isinstance(config, dict):
        for key, value in config.items():
            if not isinstance(key, str):
                raise ValueError(f"Config keys should be strings but {key} found")
            _recursive_check_config(value)

    elif isinstance(config, list):
        for value in config:
            _recursive_check_config(value)


def recursive_config_join(config1: dict, config2: dict) -> dict:
    """Recursively join 2 config objects, where config1 values override config2 values"""
    for key, value in config2.items():
        if key not in config1:
            config1[key] = value
        elif isinstance(config1[key], dict) and isinstance(value, dict):
            config1[key] = recursive_config_join(config1[key], value)

    return config1



"""
Area managers for Sentinel Hub batch grids
"""
import contextlib
import warnings
from typing import Any, List, Tuple, cast

from geopandas import GeoDataFrame
from pydantic import Field

from eolearn.core.exceptions import EODeprecationWarning
from sentinelhub import CRS, BatchRequest, BatchRequestStatus, BatchSplitter, BBox, SentinelHubBatch

from ...utils.general import convert_bbox_coords_to_int, reduce_to_coprime
from ...utils.grid import GridTransformation, create_transformations, get_grid_bbox
from .base import AreaManager


class BatchAreaManager(AreaManager):
    """Area manager that is based on Sentinel Hub Batch service tiling"""

    class Schema(AreaManager.Schema):
        """A schema used for pipelines based on batch tiling"""

        tiling_grid_id: int = Field(
            description="An id of one of the tiling grids predefined at Sentinel Hub Batch service."
        )
        resolution: float = Field(
            description=(
                "One of the resolutions that are predefined at Sentinel Hub Batch service for chosen tiling_grid_id."
            )
        )
        tile_buffer_x: int = Field(0, description="Number of pixels for which to buffer each tile left and right.")
        tile_buffer_y: int = Field(0, description="Number of pixels for which to buffer each tile up and down.")
        subsplit_x: int = Field(
            1, ge=1, description="Number of sub-tiles into which each batch tile is split along horizontal dimension."
        )
        subsplit_y: int = Field(
            1, ge=1, description="Number of sub-tiles into which each batch tile is split along vertical dimension."
        )

        batch_id: str = Field(
            "",
            description=(
                "An ID of a batch job for this pipeline. If it is given the pipeline will just monitor the "
                "existing batch job. If it is not given it will create a new batch job."
            ),
        )

    config: Schema
    _BATCH_GRID_COLUMNS = ["index_n", "id", "name", "split_x", "split_y"]
    _SH_REPROJECTION_ERROR = 1e-3  # Rounding happens in Sentinel Hub Batch database where coordinates are in WGS84

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.batch_id = self.config.batch_id

        self.subsplit = self.config.subsplit_x, self.config.subsplit_y
        self.absolute_buffer = (
            self.config.tile_buffer_x * self.config.resolution,
            self.config.tile_buffer_y * self.config.resolution,
        )

    def _get_grid_filename_params(self) -> List[object]:
        """Puts together batch parameters used for grid filename."""
        params: List[object] = [
            self.config.tiling_grid_id,
            self.config.resolution,
            self.config.tile_buffer_x,
            self.config.tile_buffer_y,
        ]
        if self.subsplit != (1, 1):
            params.extend([self.config.subsplit_x, self.config.subsplit_y])
        return params

    def _create_new_split(self, *_: Any, **__: Any) -> List[GeoDataFrame]:
        """Instead of creating a split it loads tiles from the service."""
        if not self.batch_id:
            raise MissingBatchIdError(
                "Trying to create a new batch grid but cannot collect tile geometries because 'batch_id' has not been "
                f"given. You can either define it in {self.__class__.__name__} config schema or run a pipeline that "
                "creates a new batch request."
            )

        batch_client = SentinelHubBatch(config=self.storage.sh_config)
        batch_request = batch_client.get_request(self.batch_id)
        self._verify_batch_request(batch_request)

        splitter = BatchSplitter(batch_request=batch_request, config=self.storage.sh_config)
        bbox_list = splitter.get_bbox_list()
        info_list = splitter.get_info_list()

        for info in info_list:
            info["split_x"] = 0
            info["split_y"] = 0

        if not all(bbox.crs.is_utm() for bbox in bbox_list):
            raise NotImplementedError("So far we only support UTM-based batch tiling grids")

        return self._to_dataframe_grid(bbox_list, info_list, self._BATCH_GRID_COLUMNS)

    def _verify_batch_request(self, batch_request: BatchRequest) -> None:
        """Verifies that given batch request has finished and that it has the same tiling grid parameters as
        they are written in the config.
        """
        batch_request.raise_for_status(status=[BatchRequestStatus.FAILED, BatchRequestStatus.CANCELED])

        expected_tiling_grid_params = {
            "id": self.config.tiling_grid_id,
            "bufferY": self.config.tile_buffer_y,
            "bufferX": self.config.tile_buffer_x,
            "resolution": self.config.resolution,
        }
        if batch_request.tiling_grid != expected_tiling_grid_params:
            raise ValueError(
                f"Tiling grid parameters in config are {expected_tiling_grid_params} but given batch "
                f"request has parameters {batch_request.tiling_grid}"
            )

    @staticmethod
    def _to_dataframe_grid(bbox_list: List[BBox], info_list: List[dict], info_columns: List[str]) -> List[GeoDataFrame]:
        """Sentinel Hub service provides tile info in an arbitrary order. This sorts by tile name and indices for
        consistency. Then it converts them into dataframe grid.
        """

        def sort_key_function(values: Tuple[Any, dict]) -> Tuple[str, int, int]:
            _, info = values
            return info["name"], info["split_x"], info["split_y"]

        bbox_list, info_list = zip(*sorted(zip(bbox_list, info_list), key=sort_key_function))
        for index, info_dict in enumerate(info_list):
            info_dict["index_n"] = index

        return AreaManager._to_dataframe_grid(bbox_list, info_list, info_columns)

    def transform_grid(self, target_area_manager: AreaManager) -> List[GridTransformation]:
        """Calculates how grid should be transformed into another area manager."""
        self._check_target_compatibility(target_area_manager)
        target_area_manager = cast(BatchAreaManager, target_area_manager)

        source_grid = self.get_grid(add_bbox_column=True)
        source_grid = _fix_split_columns(source_grid)
        target_grid = self._get_target_grid(target_area_manager, source_grid)

        subsplits_x, subsplits_y = zip(self.subsplit, target_area_manager.subsplit)
        source_shape, target_shape = zip(reduce_to_coprime(*subsplits_x), reduce_to_coprime(*subsplits_y))

        source_grid = self._add_patch_positions(source_grid, source_shape)
        target_grid = self._add_patch_positions(target_grid, target_shape)

        return create_transformations(source_grid, target_grid, match_columns=["name", "position_x", "position_y"])

    @staticmethod
    def _add_patch_positions(grid: List[GeoDataFrame], shape: Tuple[int, int]) -> List[GeoDataFrame]:
        """Adds columns of values that define how sub-tiles should be spatially divided into groups. Members of each
        group will have the same values in the newly added columns and will spatially form a regular subgrid of given
        shape."""
        for gdf in grid:
            gdf["position_x"] = gdf.split_x // shape[0]
            gdf["position_y"] = gdf.split_y // shape[1]

        return grid

    def _check_target_compatibility(self, target_area_manager: AreaManager) -> None:
        """Checks if source and target managers are compatible to define a transformation between them."""
        if not isinstance(target_area_manager, BatchAreaManager):
            raise NotImplementedError(
                f"Grid transformation is only supported into another {self.__class__.__name__}, but got "
                f"{target_area_manager.__class__.__name__}"
            )

        for param in ["tiling_grid_id", "resolution"]:
            if getattr(self.config, param) != getattr(target_area_manager.config, param):
                raise NotImplementedError(
                    "Cannot transform grid because source and target area managers have different values of config "
                    f"parameter {param}."
                )

    def _get_target_grid(
        self, target_area_manager: "BatchAreaManager", source_grid: List[GeoDataFrame]
    ) -> List[GeoDataFrame]:
        """First, it tries to obtain a target grid from target area manager. That would work if the grid is already
        cached or if target area manager is able to generate it. If not, then it creates the target grid from the
        given source grid."""
        with contextlib.suppress(MissingBatchIdError):
            return target_area_manager.get_grid(add_bbox_column=True)

        basic_grid = self._join_batch_grid(source_grid, self.subsplit, self.absolute_buffer)
        target_grid = self._split_batch_grid(
            basic_grid, target_area_manager.subsplit, target_area_manager.absolute_buffer
        )
        if target_area_manager.subsplit != (1, 1):
            target_grid = self._filter_grid_with_area(target_grid)

        # Bounding boxes have to be removed before caching into a Geopackage because they are not serializable.
        target_grid_without_bbox = [gdf.drop(columns=["BBOX"]) for gdf in target_grid]
        target_area_manager.cache_grid(target_grid_without_bbox)

        return target_grid

    @staticmethod
    def _join_batch_grid(
        grid: List[GeoDataFrame], split: Tuple[int, int], buffer: Tuple[float, float]
    ) -> List[GeoDataFrame]:
        """Removes buffers and joins sub-tiles from a batch grid into original tiles."""
        reversed_buffer = -buffer[0], -buffer[1]

        joined_grid = []
        for gdf in grid:
            # A single bbox is enough to reconstruct the grid tile bbox.
            reduced_gdf = gdf.groupby(by="name").first().reset_index()

            reduced_gdf.BBOX = reduced_gdf.BBOX.apply(lambda bbox: bbox.buffer(reversed_buffer, relative=False))
            reduced_gdf.BBOX = reduced_gdf.filter(["BBOX", "split_x", "split_y"]).apply(
                lambda row: get_grid_bbox(row.BBOX, (row.split_x, row.split_y), split), axis=1
            )

            # Performs a numerical correction because it assumes the base batch grid is in UTM and all coordinates are
            # aligned to integers.
            reduced_gdf.BBOX = reduced_gdf.BBOX.apply(convert_bbox_coords_to_int)

            joined_grid.append(reduced_gdf)

        return joined_grid

    def _split_batch_grid(
        self, grid: List[GeoDataFrame], split: Tuple[int, int], buffer: Tuple[float, float]
    ) -> List[GeoDataFrame]:
        """Splits tiles of batch grid into sub-tiles and applies buffer."""
        new_tiles = []
        for gdf in grid:
            for _, row in gdf.iterrows():
                tile_bbox: BBox = row.BBOX
                partition = tile_bbox.get_partition(num_x=split[0], num_y=split[1])

                for index_x, bbox_list in enumerate(partition):
                    for index_y, bbox in enumerate(bbox_list):
                        tile_info = {
                            "name": row["name"],
                            "id": row["id"],
                            "split_x": index_x,
                            "split_y": index_y,
                            "BBOX": bbox.buffer(buffer, relative=False),
                        }
                        new_tiles.append(tile_info)

        bbox_list = [item["BBOX"] for item in new_tiles]
        return self._to_dataframe_grid(bbox_list, new_tiles, self._BATCH_GRID_COLUMNS + ["BBOX"])

    def _filter_grid_with_area(self, grid: List[GeoDataFrame]) -> List[GeoDataFrame]:
        """Removes those (sub)tiles in the grid that don't intersect with the area of interest."""
        area_geometry = self.get_area_geometry(ignore_region_filter=True)

        filtered_grid = []
        for gdf in grid:
            transformed_area_geometry = area_geometry.transform(CRS(gdf.crs))
            gdf = gdf[gdf.intersects(transformed_area_geometry.geometry)]
            filtered_grid.append(gdf)

        return filtered_grid


class MissingBatchIdError(ValueError):
    """Exception that is triggered if ID of a Sentinel Hub batch job is missing."""


def _fix_split_columns(grid: List[GeoDataFrame]) -> List[GeoDataFrame]:
    """This is temporary fix for handling batch grids that were cached with older eo-grow version."""
    columns_missing = False
    for gdf in grid:
        for column_name in ["split_x", "split_y"]:
            if column_name not in gdf:
                gdf[column_name] = 0
                columns_missing = True

    if columns_missing:
        warnings.warn(
            "Since eo-grow 1.1.0 cached batch grid should also have columns split_x and split_y. "
            "For the future, make sure to clear cached grid file and reproduce it.",
            category=EODeprecationWarning,
        )

    return grid



"""
A base AreaManager implementation
"""
import logging
from collections import defaultdict
from typing import Any, List, Optional

import fiona
import fs
import geopandas as gpd
import shapely.ops
from pydantic import Field

from sentinelhub import CRS, BBox, Geometry

from ...utils.fs import LocalFile
from ...utils.grid import GridTransformation
from ...utils.types import Path
from ...utils.vector import count_points
from ..base import EOGrowObject
from ..schemas import ManagerSchema
from ..storage import StorageManager

LOGGER = logging.getLogger(__name__)


class AreaManager(EOGrowObject):
    """A class that manages AOI and how it is split into EOPatches"""

    class Schema(ManagerSchema):
        area_filename: Path
        area_buffer: Optional[float] = Field(
            description=(
                "Buffer that will be applied to AOI geometry. Buffer has to be in the same units as AOI CRS. "
                "In case buffer is too small, relatively to AOI size, it won't have any effect."
            ),
        )
        area_simplification_factor: Optional[float] = Field(
            description=(
                "A tolerance factor in CRS units how much the buffered area geometry will be simplified before "
                "splitting."
            ),
        )
        region_column_name: Optional[str] = Field(description="Name of the column in area file which contains regions.")
        region_names: Optional[List[str]] = Field(
            description="A list of regions which will be used. By default all regions are used."
        )

    config: Schema

    def __init__(self, config: Schema, storage: StorageManager):
        """
        :param config: A configuration file
        :param storage: An instance of StorageManager class
        """
        super().__init__(config)

        self.storage = storage

    def get_area_dataframe(self, *, ignore_region_filter: bool = False) -> gpd.GeoDataFrame:
        """Provides a GeoDataFrame with the AOI, which can be split over multiple sub-areas, each in a separate row.

        Current implementation is intentionally working without a LocalFile abstraction in order to be able to read
        formats consisting of multiple files (e.g. Shapefile or Geodatabase)

        :param ignore_region_filter: If `True` it will not filter by given region config parameters
        :return: A GeoDataFrame containing the unmodified area shape
        """
        filename = fs.path.join(self.storage.get_input_data_folder(), self.config.area_filename)
        with LocalFile(filename, mode="r", filesystem=self.storage.filesystem) as local_file:
            area_df = gpd.read_file(local_file.path, engine=self.storage.config.geopandas_backend)

        if self.has_region_filter() and not ignore_region_filter:
            area_df = area_df[area_df[self.config.region_column_name].isin(self.config.region_names)]

        return area_df

    def get_area_geometry(self, *, ignore_region_filter: bool = False) -> Geometry:
        """Provides a single geometry object of entire AOI

        :param ignore_region_filter: If `True` it will not filter by given region config parameters
        """
        area_filename = self._construct_file_path(prefix="area", ignore_region_filter=ignore_region_filter)

        if self.storage.filesystem.exists(area_filename):
            return self._load_area_geometry(area_filename)

        area_df = self.get_area_dataframe(ignore_region_filter=ignore_region_filter)
        area_geometry = self._process_area_geometry(area_df)

        self._save_area_geometry(area_geometry, area_filename)
        return area_geometry

    def get_grid(self, *, add_bbox_column: bool = False, ignore_region_filter: bool = False) -> List[gpd.GeoDataFrame]:
        """Provides a grid of bounding boxes which divide the AOI

        :param add_bbox_column: If `True` a new column with BBox objects will be added.
        :param ignore_region_filter: If `True` it will not filter by given region config parameters.
        :return: A list of GeoDataFrames containing bounding box geometries and their info. Bounding boxes are divided
            into GeoDataFrames per CRS.
        """
        grid_filename = self._construct_file_path(prefix="grid", ignore_region_filter=ignore_region_filter)

        if self.storage.filesystem.exists(grid_filename):
            grid = self._load_grid(grid_filename)
        else:
            grid = self._create_and_save_grid(grid_filename, ignore_region_filter)

        if add_bbox_column:
            self._add_bbox_column(grid)

        return grid

    def cache_grid(self, grid: Optional[List[gpd.GeoDataFrame]] = None, ignore_region_filter: bool = False) -> None:
        """Checks if grid is cached in the storage. If it is not, it will create and cache it.

        :param grid: If provided, this grid will be cached. Otherwise, it will generate a new grid.
        :param ignore_region_filter: If `True` it will not filter by given region config parameters.
        """
        grid_filename = self._construct_file_path(prefix="grid", ignore_region_filter=ignore_region_filter)

        if not self.storage.filesystem.exists(grid_filename):
            if grid is None:
                self._create_and_save_grid(grid_filename, ignore_region_filter)
            else:
                self._save_grid(grid, grid_filename)

    def get_grid_size(self, **kwargs: Any) -> int:
        """Calculates the number of elements of the grid

        :param kwargs: Parameters that are propagated to `get_grid` method
        :return: The number of bounding boxes in the grid
        """
        grid = self.get_grid(**kwargs)
        return sum([len(df.index) for df in grid])

    def transform_grid(self, target_area_manager: "AreaManager") -> List[GridTransformation]:
        """This method is used to define how a grid, defined by this area manager, will be transformed into a grid,
        defined by given target area manager. It calculates transformation objects between groups of bounding boxes
        from source grid and groups of bounding boxes from target grid.

        Every area manager should implement its own process of transformation and define which target area managers it
        supports."""
        raise NotImplementedError

    def has_region_filter(self) -> bool:
        """Checks region filter is set in the configuration"""
        return self.config.region_column_name is not None and self.config.region_names is not None

    def _process_area_geometry(self, area_df: gpd.GeoDataFrame) -> Geometry:
        """A method that joins area dataframe into a single geometry and applies preprocessing operations on it.

        Note: Both unary_union and buffering can take quite some time and memory when dealing with complex shapes.
        """
        LOGGER.info("Calculating a unary union of the geometries")
        area_shape = shapely.ops.unary_union(area_df.geometry)

        if self.config.area_buffer is not None:
            LOGGER.info("Applying buffer on area geometry")
            area_shape = area_shape.buffer(self.config.area_buffer)

        if self.config.area_simplification_factor is not None:
            LOGGER.info("Simplifying area geometry")
            area_shape = area_shape.simplify(self.config.area_simplification_factor, preserve_topology=True)

            point_count = count_points(area_shape)
            LOGGER.info("Simplified area shape has %d points", point_count)

        LOGGER.info("Finished processing area geometry")
        return Geometry(area_shape, CRS(area_df.crs))

    def _load_area_geometry(self, filename: str) -> Geometry:
        """Loads existing processed geometry of an area"""
        LOGGER.info("Loading cached area geometry from %s", filename)

        with LocalFile(filename, mode="r", filesystem=self.storage.filesystem) as local_file:
            area_gdf = gpd.read_file(local_file.path, engine=self.storage.config.geopandas_backend)

        return Geometry(area_gdf.geometry[0], CRS(area_gdf.crs))

    def _save_area_geometry(self, area_geometry: Geometry, filename: str) -> None:
        """Saves processed geometry of an area"""
        LOGGER.info("Saving area geometry to %s", filename)

        area_gdf = gpd.GeoDataFrame(geometry=[area_geometry.geometry], crs=area_geometry.crs.pyproj_crs())

        with LocalFile(filename, mode="w", filesystem=self.storage.filesystem) as local_file:
            area_gdf.to_file(
                local_file.path, driver="GPKG", encoding="utf-8", engine=self.storage.config.geopandas_backend
            )

    def _create_and_save_grid(self, grid_filename: str, ignore_region_filter: bool) -> List[gpd.GeoDataFrame]:
        """Defines a new grid and saves it."""
        if self.has_region_filter() and not ignore_region_filter:
            full_grid = self.get_grid(ignore_region_filter=True)
            filtered_area_geometry = self.get_area_geometry(ignore_region_filter=False)

            grid = self._filter_grid(full_grid, filtered_area_geometry)
        else:
            area_geometry = self.get_area_geometry(ignore_region_filter=ignore_region_filter)
            grid = self._create_new_split(area_geometry)

        self._save_grid(grid, grid_filename)
        return grid

    def _load_grid(self, filename: str) -> List[gpd.GeoDataFrame]:
        """A method that loads bounding box grid saved in a cache folder"""
        LOGGER.info("Loading grid from %s", filename)

        grid = []
        with LocalFile(filename, mode="r", filesystem=self.storage.filesystem) as local_file:
            for crs_layer in fiona.listlayers(local_file.path):
                grid.append(
                    gpd.read_file(local_file.path, layer=crs_layer, engine=self.storage.config.geopandas_backend)
                )

        return grid

    def _save_grid(self, grid: List[gpd.GeoDataFrame], filename: str) -> None:
        """A method that saves bounding box grid in a cache folder"""
        LOGGER.info("Saving grid to %s", filename)

        with LocalFile(filename, mode="w", filesystem=self.storage.filesystem) as local_file:
            for crs_grid in grid:
                crs_grid.to_file(
                    local_file.path,
                    driver="GPKG",
                    encoding="utf-8",
                    layer=f"Grid EPSG:{crs_grid.crs.to_epsg()}",
                    engine=self.storage.config.geopandas_backend,
                )

    def _construct_file_path(self, *, prefix: str, suffix: str = "gpkg", ignore_region_filter: bool = False) -> str:
        """Provides a file path of a cached file"""
        input_filename = self.config.area_filename.rsplit(".", 1)[0]
        input_filename = fs.path.basename(input_filename)

        filename_params: List[object] = [
            prefix,
            input_filename,
            self.__class__.__name__,
            self.config.area_buffer,
            self.config.area_simplification_factor,
        ]

        if self.has_region_filter() and not ignore_region_filter:
            filename_params.append(self.config.region_column_name)
            filename_params.extend(self.config.region_names or [])

        if prefix == "grid":
            filename_params.extend(self._get_grid_filename_params())

        filename_params = [("" if param is None else param) for param in filename_params]
        filename = "_".join(map(str, filename_params))
        filename = filename.replace(" ", "_")

        return fs.path.combine(self.storage.get_cache_folder(), f"{filename}.{suffix}")

    def _get_grid_filename_params(self) -> List[object]:
        """Provides a list of parameters to be used for grid filename. This method has to be implemented in
        subclasses.
        """
        raise NotImplementedError

    def _create_new_split(self, area_geometry: Geometry) -> List[gpd.GeoDataFrame]:
        """Splits area geometry into a grid of bounding boxes. This method has to be implemented in subclasses."""
        raise NotImplementedError

    @staticmethod
    def _to_dataframe_grid(
        bbox_list: List[BBox], info_list: List[dict], info_columns: List[str]
    ) -> List[gpd.GeoDataFrame]:
        """Splits given lists of bounding boxes and their information by CRS and for each CRS builds a GeoDataFrame

        :param bbox_list: A list of bounding boxes
        :param info_list: A list of dictionaries with information for each bounding box
        :param info_columns: A list of keys in `info_list` dictionaries that should be used to create dataframes
        """
        crs_to_bboxes = defaultdict(list)
        for bbox, info in zip(bbox_list, info_list):
            crs_to_bboxes[bbox.crs].append((bbox.geometry, info))

        dataframe_grid = []
        for crs, bbox_prop_list in crs_to_bboxes.items():
            dataframe_grid.append(
                gpd.GeoDataFrame(
                    {column: [info[column] for _, info in bbox_prop_list] for column in info_columns},
                    geometry=[geometry for geometry, _ in bbox_prop_list],
                    crs=crs.pyproj_crs(),
                )
            )
        return dataframe_grid

    @staticmethod
    def _filter_grid(grid: List[gpd.GeoDataFrame], geometry: Geometry) -> List[gpd.GeoDataFrame]:
        """Filters grid by keeping only tiles that intersect with the given geometry"""
        filtered_grid = []
        for dataframe in grid:
            projected_shape = geometry.transform(dataframe.crs).geometry
            filtered_dataframe = dataframe[dataframe.geometry.intersects(projected_shape)].copy()

            if len(filtered_dataframe.index) > 0:
                filtered_grid.append(filtered_dataframe)

        return filtered_grid

    @staticmethod
    def _add_bbox_column(grid: List[gpd.GeoDataFrame]) -> None:
        """Adds a column with bounding boxes to all dataframes in a grid"""
        for bbox_df in grid:
            crs = CRS(bbox_df.crs)
            bbox_df["BBOX"] = bbox_df.geometry.apply(lambda geo: BBox(geo.bounds, crs))  # noqa B023



"""
Module with utilities for logging
"""
import contextlib
import json
import logging
import sys
import time
from logging import FileHandler, Filter, Formatter, Handler, LogRecord, StreamHandler
from typing import Any, List, Optional, Sequence, Union

import fs
from fs.base import FS
from fs.errors import FilesystemClosed
from pydantic import Field

from eolearn.core.utils.fs import join_path, unpickle_fs

from ..utils.fs import LocalFile
from ..utils.general import jsonify
from ..utils.logging import get_instance_info
from ..utils.meta import get_package_versions
from .base import EOGrowObject
from .config import RawConfig
from .schemas import ManagerSchema
from .storage import StorageManager


class LoggingManager(EOGrowObject):
    """A class that manages logging specifics"""

    class Schema(ManagerSchema):
        save_logs: bool = Field(
            False,
            description=(
                "A flag to determine if pipeline logs and reports will be saved to "
                "logs folder. This includes potential EOExecution reports and logs."
            ),
        )
        include_logs_to_report: bool = Field(
            False,
            description=(
                "If log files should be parsed into an EOExecution report file or just linked. When working "
                "with larger number of EOPatches the recommended option is False."
            ),
        )
        eoexecution_ignore_packages: Optional[List[str]] = Field(
            description=(
                "Names of packages which logs will not be written to EOExecution log files. The default null value "
                "means that a default list of packages will be used."
            )
        )

        pipeline_ignore_packages: Optional[List[str]] = Field(
            description=(
                "Names of packages which logs will not be written to the main pipeline log file. The default null "
                "value means that a default list of packages will be used."
            )
        )
        pipeline_logs_backup_interval: float = Field(
            60,
            description=(
                "When working with a remote storage this parameter defines a minimal number of seconds between "
                "two consecutive times when pipeline log file will be copied into the remote storage."
            ),
        )

        show_logs: bool = Field(False, description="Shows basic pipeline execution logs at stdout.")
        stdout_log_packages: Optional[List[str]] = Field(
            description=(
                "Names of packages which logs will be written to stdout. The default null value means that a default "
                "list of packages will be used."
            )
        )

        capture_warnings: bool = Field(
            True,
            description=(
                "If warnings should be treated as logs and with save_logs=True written into log files instead of "
                "being printed in stderr."
            ),
        )

    config: Schema

    def __init__(self, config: Schema, storage: StorageManager):
        """
        :param config: A configuration file
        :param storage: An instance of StorageManager class
        """
        super().__init__(config)

        self.storage = storage

    def get_pipeline_logs_folder(self, pipeline_execution_name: str, full_path: bool = False) -> str:
        """Provides path to the folder where logs of this pipeline execution will be saved

        :param pipeline_execution_name: Name of current pipeline execution
        :param full_path: If it should provide a full absolute path or a path relative to the filesystem object
        """
        main_logs_folder = self.storage.get_logs_folder(full_path=full_path)
        if full_path:
            return join_path(main_logs_folder, pipeline_execution_name)
        return fs.path.combine(main_logs_folder, pipeline_execution_name)

    def start_logging(self, pipeline_execution_name: str) -> List[Handler]:
        """Creates a folder for logs and sets up (and returns) logging handlers

        Supported handlers:
        - Writing to a file in pipeline logs folder
        - Printing logs to a standard output
        """
        if self.config.save_logs:
            logs_folder = self.get_pipeline_logs_folder(pipeline_execution_name)
            self.storage.filesystem.makedirs(logs_folder, recreate=True)

        global_logger = logging.getLogger()
        global_logger.setLevel(logging.DEBUG)

        for default_handler in global_logger.handlers:
            default_handler.setLevel(logging.WARNING)

        new_handlers: List[Handler] = []

        if self.config.save_logs:
            file_handler = self._create_file_handler(pipeline_execution_name)
            global_logger.addHandler(file_handler)
            new_handlers.append(file_handler)

        if self.config.show_logs:
            stdout_handler = self._create_stdout_handler()
            global_logger.addHandler(stdout_handler)
            new_handlers.append(stdout_handler)

        if self.config.capture_warnings:
            logging.captureWarnings(True)

        return new_handlers

    def _create_file_handler(self, pipeline_execution_name: str) -> Handler:
        """Creates a logging handler to write a pipeline log to a file."""
        logs_filename = fs.path.combine(self.get_pipeline_logs_folder(pipeline_execution_name), "pipeline.log")
        file_handler = RegularBackupHandler(
            logs_filename,
            filesystem=self.storage.filesystem,
            backup_interval=self.config.pipeline_logs_backup_interval,
            encoding="utf-8",
        )

        formatter = Formatter(
            "%(levelname)s %(asctime)s %(name)s:%(lineno)d:\n\t%(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        file_handler.setFormatter(formatter)

        file_handler.addFilter(LogFileFilter(ignore_packages=self.config.pipeline_ignore_packages))

        return file_handler

    def _create_stdout_handler(self) -> Handler:
        """Creates a logging handler to write logs into a standard output."""
        stdout_handler = StreamHandler(sys.stdout)
        stdout_handler.setLevel(logging.INFO)

        formatter = Formatter("%(levelname)s %(name)s:%(lineno)d: %(message)s")
        stdout_handler.setFormatter(formatter)

        stdout_handler.addFilter(StdoutFilter(log_packages=self.config.stdout_log_packages))

        return stdout_handler

    def stop_logging(self, handlers: List[Handler]) -> None:
        """Updates logs, removes pipeline handlers from the global logger and puts global logging level back to
        default
        """
        if self.config.capture_warnings:
            logging.captureWarnings(False)

        global_logger = logging.getLogger()
        for handler in handlers:
            handler.close()
            global_logger.removeHandler(handler)

        global_logger.setLevel(logging.WARNING)

    def update_pipeline_report(
        self,
        pipeline_execution_name: str,
        pipeline_config: EOGrowObject.Schema,
        pipeline_raw_config: Optional[RawConfig],
        pipeline_id: str,
        pipeline_timestamp: str,
        elapsed_time: Optional[float] = None,
    ) -> None:
        """A method in charge of preparing a report about pipeline run.

        Content of a report:
         - pipeline configuration parameters,
         - pipeline execution stats,
         - versions of Python and Python packages,
         - information about a compute instance on which the pipeline is running.
        """
        if not self.config.save_logs:
            return

        report = {
            "config_parameters": pipeline_raw_config,
            "execution_parameters": repr(pipeline_config),
            "pipeline_execution_stats": {
                "pipeline_id": pipeline_id,
                "start_time": pipeline_timestamp,
                "elapsed_time": "<Not yet finished>" if elapsed_time is None else elapsed_time,
            },
            "versions": {"Python": sys.version, **get_package_versions()},
            "instance_info": get_instance_info(),
        }

        report_filename = fs.path.combine(
            self.get_pipeline_logs_folder(pipeline_execution_name), "pipeline-report.json"
        )
        with self.storage.filesystem.open(report_filename, "w") as report_file:
            json.dump(report, report_file, indent=2, default=jsonify)

    def save_eopatch_execution_status(self, pipeline_execution_name: str, finished: list, failed: list) -> None:
        """Saves lists of EOPatch names for which execution either finished successfully or failed"""
        if not self.config.save_logs:
            return

        filesystem = self.storage.filesystem
        logs_folder = self.get_pipeline_logs_folder(pipeline_execution_name)

        for eopatches, filename in [(finished, "finished.json"), (failed, "failed.json")]:
            path = fs.path.combine(logs_folder, filename)
            with filesystem.open(path, "w") as file:
                json.dump(eopatches, file, indent=2)


class FilesystemHandler(FileHandler):
    """A filesystem abstraction of FileHandler

    In case the handler gets a local path it behaves the same as FileHandler. In case it gets a remote path it writes
    logs first to a local path and then copies them to the remote location.

    IMPORTANT: This handler will by default have an extra `FilesystemFilter` which will ignore logs from packages that
    produce logs during `LocalFile.copy_to_remote` call. Otherwise, a log that would be created within an `emit`
    call would be recursively sent back to the handler. That would either trigger an infinite recursion or make the
    process stuck waiting for a thread lock release.
    """

    def __init__(self, path: str, filesystem: Union[FS, bytes], encoding: Optional[str] = "utf-8", **kwargs: Any):
        """
        :param path: A path to a log file that is relative to the given `filesystem` object.
        :param filesystem: A filesystem to where logs will be written. It can either be an instance of a filesystem
            object or its pickled copy.
        :param encoding: Encoding used to write log files.
        :param kwargs: Keyword arguments that will be propagated to FileHandler.
        """
        filesystem_object = unpickle_fs(filesystem) if isinstance(filesystem, bytes) else filesystem
        self.local_file = LocalFile(path, mode="w", filesystem=filesystem_object)

        super().__init__(self.local_file.path, encoding=encoding, **kwargs)

        self.addFilter(FilesystemFilter())

    def close(self) -> None:
        """Closes logging and closes the local file"""
        super().close()
        with contextlib.suppress(FilesystemClosed):
            self.local_file.close()


class RegularBackupHandler(FilesystemHandler):
    """A customized FilesystemHandler that makes a copy to a remote location regularly after given amount of time."""

    def __init__(self, path: str, filesystem: Union[FS, bytes], backup_interval: Union[float, int], **kwargs: Any):
        """
        :param path: A path to a log file that is relative to the given `filesystem` object.
        :param filesystem: A filesystem to where logs will be written. It can either be an instance of a filesystem
            object or its pickled copy.
        :param backup_interval: A minimal number of seconds before handler will back up the log file to the remote
            location. The backup will only happen when the next log record will be emitted.
        """
        super().__init__(path=path, filesystem=filesystem, **kwargs)

        self.backup_interval = backup_interval
        self._last_backup_time = time.monotonic()

    def emit(self, record: LogRecord) -> None:
        """Save a new record and backup to remote if the backup hasn't been done in the given amount of time."""
        super().emit(record)

        if time.monotonic() > self._last_backup_time + self.backup_interval:
            self.local_file.copy_to_remote()
            self._last_backup_time = time.monotonic()


class EOExecutionHandler(FilesystemHandler):
    """A customized FilesystemHandler that makes a copy to a remote location every time a new node in a workflow
    is started.
    """

    def emit(self, record: LogRecord) -> None:
        """Save a new record. In case a new node in EOWorkflow is started it will copy the log file to remote."""
        super().emit(record)

        if record.name == "eolearn.core.eoworkflow" and record.message.startswith("Computing"):
            self.local_file.copy_to_remote()


class FilesystemFilter(Filter):
    """The sole purpose of this filter is to capture any log that happens during `LocalFile.copy_to_remote` call. Any
    log that would not be captured would break the entire runtime.
    """

    IGNORE_HARMFUL_LOGS = (
        "botocore",
        "boto3.resources",
        "s3transfer",
    )

    def filter(self, record: LogRecord) -> bool:
        """Ignores logs from certain low-level packages"""
        return not record.name.startswith(self.IGNORE_HARMFUL_LOGS)


class StdoutFilter(Filter):
    """Filters log messages passed to standard output"""

    DEFAULT_LOG_PACKAGES = (
        "eogrow",
        "__main__",
        "root",
        "sentinelhub.api.batch",
    )

    def __init__(self, *args: Any, log_packages: Optional[Sequence[str]] = None, **kwargs: Any):
        """
        :param log_packages: Names of packages which logs to include.
        """
        super().__init__(*args, **kwargs)

        self.log_packages = self.DEFAULT_LOG_PACKAGES if log_packages is None else log_packages

    def filter(self, record: LogRecord) -> bool:
        """Shows only logs from eo-grow type packages and high-importance logs"""
        return any(package_name in record.name for package_name in self.log_packages)


class LogFileFilter(Filter):
    """Filters log messages passed to log file"""

    DEFAULT_IGNORE_PACKAGES = (
        "eolearn.core",
        "botocore",
        "s3transfer",
        "matplotlib",
        "fiona",
        "rasterio",
        "graphviz",
        "urllib3",
        "boto3",
    )

    def __init__(self, *args: Any, ignore_packages: Optional[Sequence[str]] = None, **kwargs: Any):
        """
        :param ignore_packages: Names of packages which logs will be ignored.
        """
        super().__init__(*args, **kwargs)

        self.ignore_packages = self.DEFAULT_IGNORE_PACKAGES if ignore_packages is None else tuple(ignore_packages)

    def filter(self, record: LogRecord) -> bool:
        """Shows everything from the main thread and process except logs from packages that are on the ignore list.
        Those packages send a lot of useless logs.
        """
        if record.name.startswith(self.ignore_packages):
            return False

        return record.threadName == "MainThread" and record.processName == "MainProcess"


class EOExecutionFilter(Filter):
    """Filters logs that will be saved by EOExecutor"""

    DEFAULT_IGNORE_PACKAGES = (
        "botocore",
        "s3transfer",
        "urllib3",
        "rasterio",
        "numba",
        "fiona.ogrext",
        "fiona.env",
        "fiona._env",
    )

    def __init__(self, ignore_packages: Optional[Sequence[str]] = None, *args: Any, **kwargs: Any):
        """
        :param ignore_packages: Names of packages which logs will be ignored.
        """
        super().__init__(*args, **kwargs)

        self.ignore_packages = self.DEFAULT_IGNORE_PACKAGES if ignore_packages is None else tuple(ignore_packages)

    def filter(self, record: LogRecord) -> bool:
        """Ignores logs from certain low-level packages"""
        return not record.name.startswith(self.ignore_packages)



""" Includes custom types used in schemas
"""
import datetime
from enum import Enum
from typing import Any, Dict, Literal, Tuple, Union

from eolearn.core import FeatureType

Path = str
S3Path = str
ImportPath = str
TimePeriod = Tuple[datetime.date, datetime.date]

Feature = Tuple[FeatureType, str]
FeatureSpec = Union[Tuple[FeatureType, str], FeatureType]

BoolOrAuto = Union[Literal["auto"], bool]

JsonDict = Dict[str, Any]
RawSchemaDict = Dict[str, Any]

AwsAclType = Literal[
    "private",
    "public-read",
    "public-read-write",
    "aws-exec-read",
    "authenticated-read",
    "bucket-owner-read",
    "bucket-owner-full-control",
    "log-delivery-write",
]


class ProcessingType(Enum):
    RAY = "ray"
    SINGLE = "single"
    MULTI = "multi"



"""
Pipelines that transform data
"""
from typing import Dict, List

from pydantic import Field

from eolearn.core import EOWorkflow, FeatureType, LoadTask, OverwritePermission, SaveTask, linearly_connect_tasks

from ..core.pipeline import Pipeline
from ..tasks.common import MappingTask
from ..utils.types import FeatureSpec


class MappingPipeline(Pipeline):
    """A pipeline for transforming a values of a mask_timeless feature according to given mapping rules"""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(description="The storage manager key pointing to the input folder")
        input_feature: str = Field(description="Name of the input mask_timeless feature")
        output_folder_key: str = Field(description="The storage manager key pointing to the output folder")
        output_feature: str = Field(description="Name of the output mask_timeless feature")

        mapping_dictionary: Dict[int, int] = Field(description="Mapping dictionary of the input-to-output classes")
        compress_level: int = Field(1, description="Level of compression used in saving eopatches")

    config: Schema

    def build_workflow(self) -> EOWorkflow:
        """Method for constructing the workflow"""
        input_feature = FeatureType.MASK_TIMELESS, self.config.input_feature
        output_feature = FeatureType.MASK_TIMELESS, self.config.output_feature

        input_features: List[FeatureSpec] = [input_feature]
        output_features: List[FeatureSpec] = [output_feature]
        if self.config.input_folder_key != self.config.output_folder_key:
            input_features.append(FeatureType.BBOX)
            output_features.append(FeatureType.BBOX)

        load_task = LoadTask(
            self.storage.get_folder(self.config.input_folder_key),
            filesystem=self.storage.filesystem,
            features=input_features,
        )

        mapping_task = MappingTask(input_feature, output_feature, self.config.mapping_dictionary)

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            features=output_features,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            compress_level=self.config.compress_level,
        )

        return EOWorkflow(linearly_connect_tasks(load_task, mapping_task, save_task))



"""
Module containing utilities for working with filesystems
"""
import abc
import os
from typing import Any, Dict, Optional, TypeVar

import fs
import fs.copy
from fs.base import FS
from fs.errors import ResourceNotFound
from fs.osfs import OSFS
from fs.tempfs import TempFS
from fs.walk import Walker

from eolearn.core.utils.fs import get_base_filesystem_and_path
from sentinelhub import SHConfig

Self = TypeVar("Self", bound="BaseLocalObject")


class BaseLocalObject(metaclass=abc.ABCMeta):
    """An abstraction for working with a local version of remote objects.

    If object's original location is remote (e.g. on S3) then this will ensure working with a local copy. If object's
    original location is already on local filesystem then it will work with that, unless `always_copy=True` is used.
    """

    def __init__(
        self,
        path: str,
        *,
        mode: str = "r",
        filesystem: Optional[FS] = None,
        config: Optional[SHConfig] = None,
        always_copy: bool = False,
        **temp_fs_kwargs: Any,
    ):
        """
        :param path: Either a full path to a remote file or a path to remote file which is relative to given filesystem
            object.
        :param mode: One of the option `r', 'w', and 'rw`, which specify if a file should be read or written to remote.
            The default is 'r'.
        :param filesystem: A filesystem of the remote. If not given, it will be determined from the path.
        :param config: A config object with which AWS credentials could be used to initialize a remote filesystem
            object.
        :param always_copy: If True it will always make a local copy to a temporary folder, even if a file is already
            in the local filesystem.
        :param temp_fs_kwargs: Parameters that will be propagated to fs.tempfs.TempFS
        """
        if filesystem is None:
            filesystem, path = get_base_filesystem_and_path(path, config=config)
        self._remote_path = path
        self._remote_filesystem = filesystem

        if not (mode and isinstance(mode, str) and set(mode) <= {"r", "w"}):
            raise ValueError(f"Parameter mode should be one of the strings 'r', 'w' or 'rw' but {mode} found")
        self._mode = mode

        if isinstance(self._remote_filesystem, (OSFS, TempFS)) and not always_copy:
            self._filesystem = self._remote_filesystem
            self._local_path = self._remote_path
        else:
            self._add_tempfs_identifier(temp_fs_kwargs)
            self._filesystem = TempFS(**temp_fs_kwargs)
            self._local_path = fs.path.basename(self._remote_path)

        self._absolute_local_path = self._filesystem.getsyspath(self._local_path)
        self._remote_location_ensured = False

        if "r" in self._mode:
            self.copy_to_local()
        if "w" in self._mode and self._filesystem is self._remote_filesystem:
            self._ensure_remote_location()

    @property
    def path(self) -> str:
        """Provides an absolute path to the copy in the local filesystem"""
        return self._absolute_local_path

    def __enter__(self: Self) -> Self:
        """This allows the class to be used as a context manager"""
        return self

    def __exit__(self, *_: Any, **__: Any) -> None:
        """This allows the class to be used as a context manager. In case an error is raised this will by default
        still delete the object from local folder.
        """
        self.close()

    def close(self) -> None:
        """Close the local copy"""
        if "w" in self._mode:
            self.copy_to_remote(raise_missing=False)

        if self._filesystem is not self._remote_filesystem:
            self._filesystem.close()

    def copy_to_local(self) -> None:
        """A public method for copying from remote to local location. It can be called anytime."""
        if self._filesystem is not self._remote_filesystem:
            self._copy_to_local()

    def copy_to_remote(self, raise_missing: bool = True) -> None:
        """Copy from local to remote location"""
        if not self._filesystem.exists(self._local_path):
            if raise_missing:
                raise ResourceNotFound(f"Local resource {self._local_path} doesn't exist")
            return

        if self._filesystem is not self._remote_filesystem:
            self._ensure_remote_location()
            self._copy_to_remote()

    def _add_tempfs_identifier(self, temp_fs_kwargs: Dict[str, Any]) -> None:
        """Adds an identifier name that will be used as a suffix of a temporary local folder. This is helpful for
        debugging purposes."""
        if "identifier" in temp_fs_kwargs:
            return

        object_name = fs.path.basename(self._remote_path.rstrip("/")).split(".", 1)[0]
        temp_fs_kwargs["identifier"] = f"_{self.__class__.__name__}-{object_name}"

    def _ensure_remote_location(self) -> None:
        """Makes sure that the remote location exists. If it doesn't then it will try to create the missing folders.
        This method is also regulated with a flag so that the IO checks happen at most once."""
        if not self._remote_location_ensured:
            remote_dirs = fs.path.dirname(self._remote_path)
            self._remote_filesystem.makedirs(remote_dirs, recreate=True)
            self._remote_location_ensured = True

    @abc.abstractmethod
    def _copy_to_local(self) -> None:
        """Copy from remote to local location"""

    @abc.abstractmethod
    def _copy_to_remote(self) -> None:
        """Copy from local to remote location"""


class LocalFile(BaseLocalObject):
    """An abstraction for working with a local version of a remote file.

    Check `BaseLocalObject` for more info.
    """

    def _copy_to_local(self) -> None:
        """Copy the file from remote to local location."""
        fs.copy.copy_file(self._remote_filesystem, self._remote_path, self._filesystem, self._local_path)

    def _copy_to_remote(self) -> None:
        """Copy the file from local to remote location."""
        fs.copy.copy_file(self._filesystem, self._local_path, self._remote_filesystem, self._remote_path)


class LocalFolder(BaseLocalObject):
    """An abstraction for working with a local version of a remote folder and its content.

    Check `BaseLocalObject` for more info.
    """

    def __init__(
        self,
        path: str,
        *args: Any,
        filesystem: Optional[FS] = None,
        config: Optional[SHConfig] = None,
        walker: Optional[Walker] = None,
        workers: Optional[int] = None,
        **kwargs: Any,
    ):
        """
        :param path: Either a full path to a remote folder or a path to remote folder which is relative to given
            filesystem object.
        :param args: Positional arguments propagated to the base class.
        :param filesystem: A filesystem of the remote. If not given, it will be determined from the path.
        :param config: A config object with which AWS credentials could be used to initialize a remote filesystem
            object.
        :param walker: An instance of `fs.walk.Walker` object used to configure advanced copying parameters. It is
            possible to set max folder depth of copy (default is entire tree), how to handle errors (by default are
            ignored), and what to include or exclude.
        :param workers: A maximal number of threads used for parallel copies between local and remote locations. The
            default is `5` times the number of CPUs.
        :param kwargs: Keyword arguments propagated to the base class.
        """
        self.walker = walker or Walker(ignore_errors=True)
        self.workers = 5 * (os.cpu_count() or 1) if workers is None else workers

        if filesystem is None:
            filesystem, path = get_base_filesystem_and_path(path, config=config)
        if not path.endswith("/"):
            path += "/"  # This way fs will treat the path as a folder

        super().__init__(path, *args, filesystem=filesystem, **kwargs)

    def _copy_to_local(self) -> None:
        """Copy the folder content from remote to local location."""
        fs.copy.copy_dir(
            self._remote_filesystem,
            self._remote_path,
            self._filesystem,
            self._local_path,
            walker=self.walker,
            workers=self.workers,
        )

    def _copy_to_remote(self) -> None:
        """Copy the folder content from local to remote location."""
        fs.copy.copy_dir(
            self._filesystem,
            self._local_path,
            self._remote_filesystem,
            self._remote_path,
            walker=self.walker,
            workers=self.workers,
        )



"""
Module implementing merge samples
"""
import logging
from typing import List, Literal, Optional, Tuple, cast

import fs
import numpy as np
from pydantic import Field

from eolearn.core import EOPatch, EOWorkflow, FeatureType, LoadTask, OutputTask, linearly_connect_tasks
from eolearn.core.utils.fs import get_full_path

from ..core.pipeline import Pipeline
from ..utils.types import Feature, FeatureSpec

LOGGER = logging.getLogger(__name__)


class MergeSamplesPipeline(Pipeline):
    """Pipeline to merge sampled data into joined numpy arrays"""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(
            description="The storage manager key pointing to the input folder for the merge samples."
        )
        output_folder_key: str = Field(
            description="The storage manager key pointing to the output folder for the merge samples pipeline."
        )
        features_to_merge: List[Feature] = Field(
            description="Dictionary of all features for which samples are to be merged."
        )
        include_timestamp: bool = Field(False, description="Whether to also prepare an array of merged timestamps.")
        id_filename: Optional[str] = Field(description="Filename of array holding patch id of concatenated features")
        suffix: str = Field("", description="String to append to array filenames")
        workers: int = Field(1, description="Number of threads used to load data from EOPatches in parallel.")
        use_ray: Literal[False] = Field(False, description="Pipeline does not parallelize properly.")
        skip_existing: Literal[False] = False

    config: Schema

    _OUTPUT_NAME = "features-to-merge"

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Procedure which merges data from EOPatches into ML-ready numpy arrays"""
        workflow = self.build_workflow()
        exec_args = self.get_execution_arguments(workflow)

        # It doesn't make sense to parallelize loading over a cluster, but it would # make sense to parallelize over
        # features that have to be concatenated or, if we would concatenate into multiple files, parallelize creating
        # batches of features
        successful, failed, results = self.run_execution(workflow, exec_args, multiprocess=False)

        result_patches = [cast(EOPatch, result.outputs.get(self._OUTPUT_NAME)) for result in results]

        self.merge_and_save_features(result_patches, patch_names=successful)

        return successful, failed

    def build_workflow(self) -> EOWorkflow:
        """Creates a workflow that outputs the requested features"""
        features_to_load: List[FeatureSpec] = [FeatureType.TIMESTAMP] if self.config.include_timestamp else []
        features_to_load.extend(self.config.features_to_merge)
        load_task = LoadTask(
            self.storage.get_folder(self.config.input_folder_key),
            filesystem=self.storage.filesystem,
            features=features_to_load,
        )
        output_task = OutputTask(name=self._OUTPUT_NAME)
        return EOWorkflow(linearly_connect_tasks(load_task, output_task))

    def merge_and_save_features(self, patches: List[EOPatch], patch_names: List[str]) -> None:
        """Merges features from EOPatches and saves data"""
        patch_sample_nums = None

        for feature in self.config.features_to_merge:
            LOGGER.info("Started merging feature %s", feature)
            arrays = [self._collect_and_remove_feature(patch, feature) for patch in patches]
            _, feature_name = feature

            if patch_sample_nums is None:
                patch_sample_nums = [array.shape[0] for array in arrays]

            merged_array: np.ndarray = np.concatenate(arrays, axis=0)
            del arrays

            self._save_array(merged_array, feature_name)
            del merged_array

        if patch_sample_nums is None:
            raise ValueError("Need at least one feature to merge.")

        if self.config.include_timestamp:
            arrays = []
            for patch, sample_num in zip(patches, patch_sample_nums):
                arrays.append(np.tile(np.array(patch.timestamp), (sample_num, 1)))
                patch.timestamp = []

            self._save_array(np.concatenate(arrays, axis=0), "TIMESTAMPS")

        if self.config.id_filename:
            LOGGER.info("Started merging EOPatch ids")
            patch_ids = self.eopatch_manager.get_id_list_from_eopatch_list(patch_names)
            patch_id_arrays = [
                np.ones(sample_num, dtype=np.uint32) * patch_id
                for sample_num, patch_id in zip(patch_sample_nums, patch_ids)
            ]

            merged_patch_ids: np.ndarray = np.concatenate(patch_id_arrays, axis=0)
            self._save_array(merged_patch_ids, self.config.id_filename)

    @staticmethod
    def _collect_and_remove_feature(patch: EOPatch, feature: Feature) -> np.ndarray:
        """Collects a feature from an EOPatch and removes it from EOPatch to conserve overall memory"""
        feature_array = patch[feature]
        feature_type, _ = feature

        if feature_type is FeatureType.TIMESTAMP:
            patch.timestamp = []
            return np.array(feature_array)

        del patch[feature]

        axis = feature_type.ndim() - 2  # type: ignore[operator]
        feature_array = np.squeeze(feature_array, axis=axis)

        if feature_type in [FeatureType.DATA, FeatureType.MASK]:
            feature_array = np.moveaxis(feature_array, 0, 1)

        if feature_array.dtype == np.float64:
            feature_array = feature_array.astype(np.float32)

        return feature_array

    def _save_array(self, array: np.ndarray, name: str) -> None:
        """Saves an array to the storage with training data using the name of the array"""
        save_folder = self.storage.get_folder(self.config.output_folder_key)
        if self.config.suffix:
            name = f"{name}_{self.config.suffix}"
        path = fs.path.combine(save_folder, f"{name}.npy")

        with self.storage.filesystem.openbin(path, "w") as file_handle:
            np.save(file_handle, array)

        LOGGER.info("Saved concatenated array to %s", get_full_path(self.storage.filesystem, path))



"""
Utilities for filtering eopatch lists
"""
from concurrent.futures import ThreadPoolExecutor
from typing import List, Sequence, Set, Tuple, Union

import fs
from fs.base import FS
from tqdm.auto import tqdm

from eolearn.core import FeatureType
from eolearn.core.eodata_io import walk_filesystem
from eolearn.core.utils.types import EllipsisType

from ..utils.types import FeatureSpec


def check_if_features_exist(
    filesystem: FS,
    eopatch_path: str,
    features: Sequence[FeatureSpec],
) -> bool:
    """Checks whether an EOPatch in the given location has all specified features saved"""
    not_seen_features: Set[Union[FeatureType, Tuple[FeatureType, Union[str, EllipsisType]]]] = set(features)
    try:
        for ftype, name, _ in walk_filesystem(filesystem, eopatch_path, features=features):
            if (ftype, name) in not_seen_features:
                not_seen_features.remove((ftype, name))
            elif ftype in not_seen_features:
                not_seen_features.remove(ftype)
    except (IOError, fs.errors.ResourceNotFound):
        return False
    return len(not_seen_features) == 0


def get_patches_with_missing_features(
    filesystem: FS,
    patches_folder: str,
    patch_list: List[str],
    features: Sequence[FeatureSpec],
) -> List[str]:
    """Filters out names of those EOPatches that are missing some given features.

    :param filesystem: A filesystem object.
    :param patches_folder: A path to folder with EOPatches, relative to `filesystem` object.
    :param patch_list: A list of EOPatch names.
    :param features: A list of EOPatch features.
    :return: A sublist of `patch_list` with only EOPatch names that are missing some features.
    """
    eopatch_paths = [fs.path.combine(patches_folder, eopatch) for eopatch in patch_list]

    def check_patch(eopatch_path: str) -> bool:
        return check_if_features_exist(filesystem, eopatch_path, features)

    with ThreadPoolExecutor() as executor:
        has_features_list = list(
            tqdm(
                executor.map(check_patch, eopatch_paths),
                total=len(eopatch_paths),
                desc="Checking EOPatches",
            )
        )

    return [eopatch for eopatch, has_features in zip(patch_list, has_features_list) if not has_features]



"""
Definition of tasks needed for calculating features
"""
import abc
from datetime import date, datetime, time, timedelta
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np

from eolearn.core import EOPatch, EOTask, FeatureType, MapFeatureTask

from ..utils.types import Feature


def join_valid_and_cloud_masks(valid_mask: np.ndarray, cloud_mask: np.ndarray) -> np.ndarray:
    """Used to zip together information about valid data and clouds into a combined validity mask"""
    return valid_mask.astype(bool) & (cloud_mask == 0)


class ValidDataFractionPredicate:
    """
    Predicate that defines if a frame from EOPatch's time-series is valid or not. Frame is valid, if the
    valid data fraction is above the specified threshold.
    """

    def __init__(self, validity_threshold: float):
        self.validity_threshold = validity_threshold

    def __call__(self, array: np.ndarray) -> bool:
        coverage = np.sum(array.astype(np.uint8)) / np.prod(array.shape)
        return (coverage > self.validity_threshold).astype(bool)


class MaxNDVI(MapFeatureTask):
    def map_method(self, feature: np.ndarray) -> np.ndarray:
        if feature.shape[0]:
            return np.nanmax(feature, axis=0)
        # A special case of arrays with time dimension of size 0
        return np.full(feature.shape[1:], np.nan, dtype=feature.dtype)


class MosaickingTask(EOTask, metaclass=abc.ABCMeta):
    """Base class for mosaicking images given an interval of edge dates"""

    def __init__(
        self,
        feature: Feature,
        dates: Union[List[date], Tuple[date, date, int]],
        valid_mask: Optional[Feature] = None,
        ndvi_feature: Optional[Feature] = None,
    ):
        self.parsed_feature = self.parse_renamed_feature(feature, allowed_feature_types={FeatureType.DATA})
        self.valid_mask_type, self.valid_mask_name = None, None
        if valid_mask is not None:
            self.valid_mask_type, self.valid_mask_name = self.parse_feature(
                valid_mask, allowed_feature_types={FeatureType.MASK}
            )
        self.ndvi_feature_type, self.ndvi_feature_name = None, None
        if ndvi_feature is not None:
            self.ndvi_feature_type, self.ndvi_feature_name = self.parse_feature(
                ndvi_feature, allowed_feature_types={FeatureType.DATA}
            )
        self.dates = self._get_dates(dates)

    def _get_dates(self, dates: Union[List[date], Tuple[date, date, int]]) -> np.ndarray:
        """Set dates either from list of dates or a tuple (start_date, end_date, n_mosaics)"""
        if all(isinstance(d, (date, datetime)) for d in dates):
            return np.array(dates)
        if len(dates) == 3 and isinstance(dates[-1], int):
            return self._get_date_edges(*dates)
        raise ValueError(
            "dates parameter can be either a list of date(time)s or a tuple "
            "(start_date, end_date, n_mosaics) for equidistant intervals between start and end date."
        )

    @staticmethod
    def _get_date_edges(start_date: date, end_date: date, parts: int) -> np.ndarray:
        """Help function to get dates of year split into equal parts

        :param start_date: first date of time interval
        :param end_date: last date of time interval
        :param parts: Number of parts to split the year into
        :return: numpy array of dates that split the time interval into equal parts
        """
        start = datetime.combine(start_date, time.min)
        end = datetime.combine(end_date, time.min) + timedelta(days=1)
        diff = (end - start) / parts
        edges = [start + diff * i for i in range(parts)]
        edges.append(end)
        return np.array(edges)

    def _find_time_indices(self, timestamps: Sequence[date], index: int) -> Tuple[np.ndarray, ...]:
        """Compute indices of images to use for mosaicking"""
        if index == 1:
            array = np.where((np.array(timestamps) <= self.dates[index]))
        elif index == len(self.dates) - 1:
            array = np.where((np.array(timestamps) > self.dates[index - 1]))
        else:
            array = np.where(
                (np.array(timestamps) > self.dates[index - 1]) & (np.array(timestamps) <= self.dates[index])
            )
        return array

    def compute_mosaic_dates(self) -> np.ndarray:
        """Compute dates of corresponding mosaics"""
        # calculate centers of date edges
        delta = self.dates[1:] - self.dates[:-1]
        return self.dates[:-1] + delta / 2

    @abc.abstractmethod
    def _compute_single_mosaic(self, eopatch: EOPatch, idate: int) -> np.ndarray:
        """Compute single mosaic given index of edge date"""

    def compute_mosaic(self, eopatch: EOPatch) -> np.ndarray:
        """Computes mosaic"""
        return np.array([self._compute_single_mosaic(eopatch, idate) for idate in range(1, len(self.dates))])

    def execute(self, eopatch: EOPatch) -> EOPatch:
        """Compute mosaic for given dates"""
        feature_type, _, new_feature_name = self.parsed_feature

        eopatch.timestamp = [ts.replace(tzinfo=None) for ts in eopatch.timestamp]
        eopatch[(feature_type, new_feature_name)] = self.compute_mosaic(eopatch)

        eopatch.timestamp = list(self.compute_mosaic_dates())

        return eopatch


class MaxNDVIMosaickingTask(MosaickingTask):
    """
    Task to create mosaics of data based on the max NDVI value between provided dates
    """

    def __init__(
        self,
        feature: Feature,
        dates: Union[List[date], Tuple[date, date, int]],
        ndvi_feature: Feature,
        valid_mask: Optional[Feature] = None,
    ):
        super().__init__(feature, dates, ndvi_feature=ndvi_feature, valid_mask=valid_mask)

    def _compute_single_mosaic(self, eopatch: EOPatch, idate: int) -> np.ndarray:
        """Compute single mosaic using values of the max NDVI"""
        array = self._find_time_indices(eopatch.timestamp, idate)
        feature_type, feature_name, _ = self.parsed_feature
        feat_values = eopatch[(feature_type, feature_name)][array].astype(np.float32)
        ndvi_values = eopatch[(self.ndvi_feature_type, self.ndvi_feature_name)][array]  # type: ignore[index]
        valid_mask = (
            eopatch[self.valid_mask_type][self.valid_mask_name][array]
            if self.valid_mask_type is not None
            else np.ones(feat_values.shape, dtype=bool)
        ).astype(bool)

        ndvi_values[~valid_mask] = np.nan
        feat_values[~np.broadcast_to(valid_mask, feat_values.shape)] = np.nan

        mask_nan_slices = np.all(np.isnan(ndvi_values), axis=0, keepdims=True)
        ndvi_values[np.broadcast_to(mask_nan_slices, ndvi_values.shape)] = -999
        feat_values[np.broadcast_to(mask_nan_slices, feat_values.shape)] = -999

        timeframes, height, width, depth = feat_values.shape

        if timeframes == 0:
            mosaic = np.full((height, width, depth), np.nan)
        else:
            if timeframes == 1:
                mosaic = feat_values[0]
            else:
                indices = np.nanargmax(ndvi_values, axis=0).squeeze(axis=-1)
                ixgrid: Tuple[np.ndarray, ...] = np.ix_(np.arange(timeframes), np.arange(height), np.arange(width))
                mosaic = feat_values[indices, ixgrid[1], ixgrid[2], :].squeeze(axis=0)

            mosaic[np.broadcast_to(mask_nan_slices[0], mosaic.shape)] = np.nan
        return mosaic


class MedianMosaickingTask(MosaickingTask):
    """
    Task to create mosaics of data based on the median value between provided dates
    """

    def __init__(
        self,
        feature: Feature,
        dates: Union[List[date], Tuple[date, date, int]],
        valid_mask: Optional[Feature] = None,
    ):
        super().__init__(feature, dates, valid_mask=valid_mask)

    def _compute_single_mosaic(self, eopatch: EOPatch, idate: int) -> np.ndarray:
        """Compute single mosaic using the median of values"""
        array = self._find_time_indices(eopatch.timestamp, idate)
        feature_type, feature_name, _ = self.parsed_feature

        feat_values = eopatch[(feature_type, feature_name)][array].astype(np.float32)
        valid_mask = (
            eopatch[(self.valid_mask_type, self.valid_mask_name)][array]
            if self.valid_mask_type is not None
            else np.ones(feat_values.shape, dtype=bool)
        )

        feat_values[~np.broadcast_to(valid_mask, feat_values.shape)] = np.nan

        mask_nan_slices = np.all(np.isnan(feat_values), axis=0, keepdims=True)
        feat_values[np.broadcast_to(mask_nan_slices, feat_values.shape)] = -999
        timeframes, height, width, depth = feat_values.shape

        if timeframes == 0:
            mosaic = np.full((height, width, depth), np.nan)
        else:
            if timeframes == 1:
                mosaic = feat_values[0]
            else:
                mosaic = np.nanmedian(feat_values, axis=0)

            mosaic[np.broadcast_to(mask_nan_slices[0], mosaic.shape)] = np.nan

        return mosaic



"""
Tasks for spatial operations on EOPatches
"""
from typing import Dict, List, Optional, Tuple, Union, cast

import numpy as np
from geopandas import GeoDataFrame

from eolearn.core import EOPatch, EOTask, FeatureType, deep_eq
from sentinelhub import CRS, BBox, bbox_to_resolution

from ..utils.general import convert_to_int
from ..utils.types import Feature, FeatureSpec
from ..utils.vector import concat_gdf


class SpatialJoinTask(EOTask):
    """Spatially joins data from multiple EOPatches"""

    ERR = 1e-8

    def __init__(
        self,
        features: List[FeatureSpec],
        no_data_map: Dict[Feature, float],
        unique_columns_map: Dict[Feature, List[str]],
        raise_misaligned: bool = True,
    ):
        self.features = self.parse_features(features)
        self.no_data_map = no_data_map
        self.unique_columns_map = unique_columns_map
        self.raise_misaligned = raise_misaligned

    def _join_spatial_rasters(
        self, rasters: List[np.ndarray], bboxes: List[BBox], joined_bbox: BBox, no_data_value: float
    ) -> np.ndarray:
        """Joins all rasters into a single new rasters."""
        if len({raster.dtype for raster in rasters}) != 1:
            raise ValueError("Cannot join raster features with different dtypes")

        resolution = self._get_resolution(rasters, bboxes)
        joined_raster = self._get_joined_raster(joined_bbox, resolution, rasters[0], no_data_value)
        height, width = joined_raster.shape[-3:-1]

        for raster, bbox in zip(rasters, bboxes):
            slice_x, slice_y = get_array_slices(
                joined_bbox,
                bbox,
                resolution=resolution,
                raise_misaligned=self.raise_misaligned,
                limit_x=(0, width),
                limit_y=(0, height),
            )
            joined_raster[..., slice_y, slice_x, :] = raster

        return joined_raster

    def _get_resolution(self, rasters: List[np.ndarray], bboxes: List[BBox]) -> np.ndarray:
        """Checks that resolutions of given geospatial rasters don't differ more than a rounding error."""
        resolutions = []

        for raster, bbox in zip(rasters, bboxes):
            height, width = raster.shape[-3:-1]
            resolutions.append(np.array(bbox_to_resolution(bbox, width=width, height=height, meters=False)))

        mean_resolution = np.mean(resolutions, axis=0)
        for resolution in resolutions:
            diff = np.amax(np.absolute(resolution - mean_resolution))
            if diff > self.ERR:
                raise ValueError("Rasters have different spatial resolutions therefore they cannot be joined together")

        return mean_resolution

    def _get_joined_raster(
        self, bbox: BBox, resolution: np.ndarray, sample_raster: np.ndarray, no_data_value: float
    ) -> np.ndarray:
        """Provides an empty raster for a given bbox and resolution."""
        image_size = np.abs(np.array(bbox.upper_right) - np.array(bbox.lower_left)) / resolution
        try:
            width, height = convert_to_int(image_size, raise_diff=self.raise_misaligned)
        except ValueError as exception:
            raise ValueError(
                "Pixels from joined raster would be misaligned with pixels from original rasters"
            ) from exception

        time_dim = sample_raster.shape[:-3]  # Either empty or with a single element
        shape = (*time_dim, height, width, sample_raster.shape[-1])
        return np.full(shape, no_data_value, dtype=sample_raster.dtype)

    @staticmethod
    def _join_vector_data(dataframes: List[GeoDataFrame], unique_columns: Optional[List[str]]) -> GeoDataFrame:
        """Joins dataframes and optionally drops duplicated rows."""
        joined_dataframe = concat_gdf(dataframes)
        if unique_columns:
            return joined_dataframe.drop_duplicates(subset=unique_columns)
        return joined_dataframe

    def execute(self, *eopatches: EOPatch, bbox: BBox) -> EOPatch:
        """Spatially joins given EOPatches into a new EOPatch with given bounding box."""
        eopatches = tuple(eopatch for eopatch in eopatches if eopatch.get_features())
        for eopatch in eopatches:
            if eopatch.bbox is None:
                raise ValueError("All non-empty input EOPatches should have a bounding box")
            if eopatch.bbox.crs is not bbox.crs:
                raise ValueError("EOPatches must have the same CRS as the given bounding box")

        # Sorting EOPatches in unique order so that the values in the overlapping areas will always be computed in
        # the same way.
        eopatches = tuple(sorted(eopatches, key=lambda eop: repr(eop.bbox)))

        joined_eopatch = EOPatch(bbox=bbox)

        for feature in self.features:
            feature_type, _ = feature
            if feature_type is FeatureType.BBOX:
                continue

            data = [eopatch[feature] for eopatch in eopatches if feature in eopatch]
            if not data:
                continue

            if feature_type.is_spatial():
                feature = cast(Feature, feature)  # bbox and timestamp are discarded with above check
                if feature_type.is_raster():
                    bboxes: List[BBox] = [patch.bbox for patch in eopatches if feature in patch]  # type: ignore[misc]
                    joined_data = self._join_spatial_rasters(data, bboxes, bbox, self.no_data_map[feature])
                else:
                    joined_data = self._join_vector_data(data, self.unique_columns_map.get(feature))
            else:
                joined_data = data[0]

                if not all(deep_eq(joined_data, item) for item in data[1:]):
                    raise ValueError(
                        f"Different EOPatches have different values of a non-spatial feature {feature}. It is not clear"
                        " how to join them."
                    )

            joined_eopatch[feature] = joined_data

        return joined_eopatch


class SpatialSliceTask(EOTask):
    """Spatially slices given EOPatch to create a new one."""

    def __init__(self, features: List[FeatureSpec], raise_misaligned: bool = True):
        self.features = self.parse_features(features)
        self.raise_misaligned = raise_misaligned

    def _slice_raster(self, raster: np.ndarray, raster_bbox: BBox, slice_bbox: BBox) -> np.ndarray:
        """Spatially slice a raster array."""
        height, width = raster.shape[-3:-1]
        slice_x, slice_y = get_array_slices(
            raster_bbox, slice_bbox, size=(width, height), raise_misaligned=self.raise_misaligned
        )
        return raster[..., slice_y, slice_x, :]

    @staticmethod
    def _filter_vector(gdf: GeoDataFrame, bbox: BBox) -> GeoDataFrame:
        """Spatially filters a GeoDataFrame."""
        bbox = bbox.transform(CRS(gdf.crs))
        intersects_bbox = gdf.geometry.intersects(bbox.geometry)
        return gdf[intersects_bbox].copy(deep=True)

    def execute(self, eopatch: EOPatch, *, bbox: Optional[BBox]) -> EOPatch:
        """Spatially slices given EOPatch with given bounding box.

        The task allows that no bounding box for slicing is provided. In such case the task will return an EOPatch with
        only non-spatial features.
        """
        main_bbox = eopatch.bbox
        if not main_bbox:
            raise ValueError("EOPatch is missing a bounding box")
        if bbox is not None:
            if main_bbox.crs is not bbox.crs:
                raise ValueError("Given bbox is not in the same CRS as EOPatch bbox")
            if not main_bbox.geometry.contains(bbox.geometry):
                raise ValueError("Given bbox must be fully contained in EOPatch's bbox")

        sliced_eopatch = EOPatch(bbox=bbox)

        for feature in self.features:
            feature_type, _ = feature
            if feature_type is FeatureType.BBOX:
                continue
            data = eopatch[feature]

            if feature_type.is_spatial():
                if bbox is None:
                    continue

                if feature_type.is_raster():
                    sliced_data = self._slice_raster(data, main_bbox, bbox)
                else:
                    sliced_data = self._filter_vector(data, bbox)
            else:
                sliced_data = data

            sliced_eopatch[feature] = sliced_data

        return sliced_eopatch


def get_array_slices(
    bbox: BBox,
    slice_bbox: BBox,
    *,
    resolution: Optional[Union[np.ndarray, Tuple[float, float]]] = None,
    size: Optional[Union[np.ndarray, Tuple[int, int]]] = None,
    raise_misaligned: bool = True,
    limit_x: Optional[Tuple[int, int]] = None,
    limit_y: Optional[Tuple[int, int]] = None,
) -> Tuple[slice, slice]:
    """Slicing taken from eolearn.io.ImportFromTiffTask.

    :param bbox: A bounding box of initial array.
    :param slice_bbox: A bounding box of array to be sliced.
    :param resolution: A working resolution.
    :param size: A working size.
    :param raise_misaligned: Whether to raise an error if the slice would be pixel misaligned the initial array.
    :param limit_x: If provided it will clip the horizontal slice to a given interval.
    :param limit_y: If provided it will clip the vertical slice to a given interval.
    :return: A slice over horizontal direction and a slice over vertical direction.
    """
    raster_upper_left = np.array([bbox.min_x, bbox.max_y])

    if size is not None and resolution is None:
        width, height = size
        raster_lower_right = np.array([bbox.max_x, bbox.min_y])
        resolution = np.abs((raster_upper_left - raster_lower_right)) / (width, height)
    elif size is None and resolution is not None:
        resolution = np.array(resolution)
    else:
        raise ValueError("Only one of the resolution and size can be given")

    slice_upper_left = np.array([slice_bbox.min_x, slice_bbox.max_y])
    slice_lower_right = np.array([slice_bbox.max_x, slice_bbox.min_y])
    axis_flip = [1, -1]  # image origin is upper left, geographic origin is lower left

    offset = axis_flip * (slice_upper_left - raster_upper_left) / resolution
    try:
        offset_x, offset_y = convert_to_int(offset, raise_diff=raise_misaligned)
        slice_size = abs(slice_lower_right - slice_upper_left) / resolution
        slice_width, slice_height = convert_to_int(slice_size, raise_diff=raise_misaligned)
    except ValueError as exception:
        raise ValueError(
            "Pixels from a slice raster would be misaligned with pixels from an original raster"
        ) from exception

    interval_x = offset_x, offset_x + slice_width
    if limit_x:
        interval_x = max(interval_x[0], limit_x[0]), min(interval_x[1], limit_x[1])
    interval_y = offset_y, offset_y + slice_height
    if limit_y:
        interval_y = max(interval_y[0], limit_y[0]), min(interval_y[1], limit_y[1])

    return slice(*interval_x), slice(*interval_y)



"""
Utilities for working with area grids
"""
from collections import defaultdict
from dataclasses import dataclass, field
from typing import DefaultDict, List, Sequence, Tuple

import numpy as np
from geopandas import GeoDataFrame
from pandas import DataFrame
from shapely.geometry import GeometryCollection

from sentinelhub import BBox


@dataclass
class GridTransformation:
    """A dataclass holding information about a transformation from a group of source bounding boxes into a group of
    target bounding boxes via an enclosing bounding box.

    The dataclass also holds a source and a target dataframe which contain any additional information about bounding
    boxes that would be required to use at any later stage.
    """

    enclosing_bbox: BBox
    source_bboxes: Tuple[BBox, ...]
    target_bboxes: Tuple[BBox, ...]
    source_df: DataFrame = field(default_factory=DataFrame)
    target_df: DataFrame = field(default_factory=DataFrame)

    def __post_init__(self) -> None:
        if not self.target_bboxes:
            raise ValueError("There should be at least 1 target bounding box to make this transformation valid.")


def create_transformations(
    source_grid: List[GeoDataFrame], target_grid: List[GeoDataFrame], match_columns: List[str]
) -> List[GridTransformation]:
    """Given a source and a target dataframe it splits them into groups. Items in each group have to have the same
    values in `match_columns` columns. Then it creates a transformation object for each group."""
    transformations = []
    crs_map = {gdf.crs: gdf for gdf in source_grid}
    for target_gdf in target_grid:
        source_gdf = crs_map.get(target_gdf.crs)

        target_map = dict(iter(target_gdf.groupby(by=match_columns)))

        source_map: DefaultDict[tuple, DataFrame] = defaultdict(DataFrame)
        if source_gdf is not None:
            source_map = defaultdict(DataFrame, iter(source_gdf.groupby(by=match_columns)))

        for group_values, target_group in target_map.items():
            source_group = source_map[group_values]

            source_bboxes = source_group.BBOX.tolist() if len(source_group) else []
            target_bboxes = target_group.BBOX.tolist()
            enclosing_bbox = get_enclosing_bbox(source_bboxes + target_bboxes)

            transformation = GridTransformation(
                enclosing_bbox=enclosing_bbox,
                source_bboxes=tuple(source_bboxes),
                target_bboxes=tuple(target_bboxes),
                source_df=source_group,
                target_df=target_group,
            )
            transformations.append(transformation)

    return transformations


def get_grid_bbox(bbox: BBox, index: Tuple[int, int], split: Tuple[int, int]) -> BBox:
    """For a bounding box from a regular grid of bounding boxes, its position in the grid and the split of the grid it
    calculates the bounding box of the grid.

    :param bbox: A single bounding box from a regular grid. Bounding box shouldn't have any buffer.
    :param index: A pair of horizontal and vertical indices of a bounding box in a grid.
    :param split: A pair of numbers of splits in horizontal and vertical direction.
    :return: A total bounding box of the grid.
    """
    lower_left = np.array(bbox.lower_left)
    upper_right = np.array(bbox.upper_right)

    size = upper_right - lower_left
    grid_lower_left = lower_left - index * size
    grid_upper_right = lower_left + split * size - index * size

    return BBox((*grid_lower_left, *grid_upper_right), crs=bbox.crs)


def get_enclosing_bbox(bboxes: Sequence[BBox]) -> BBox:
    """A minimal bounding box that encloses all given bounding boxes"""
    unique_crs = {bbox.crs for bbox in bboxes}
    if len(unique_crs) != 1:
        raise ValueError(f"Given bounding boxes should have a single unique CRS, found {unique_crs}")

    collection = GeometryCollection([bbox.geometry for bbox in bboxes])
    return BBox(collection.bounds, crs=bboxes[0].crs)



"""
For working with pre-defined grids
"""
import logging
from typing import Any, List

import fs
import geopandas as gpd
import shapely.ops
from pydantic import Field

from sentinelhub import CRS, Geometry

from ...utils.types import Path
from ...utils.vector import concat_gdf
from ..schemas import ManagerSchema
from .base import AreaManager

LOGGER = logging.getLogger(__name__)


class CustomGridAreaManager(AreaManager):
    """Area manager that works with a pre-defined grid of EOPatches"""

    class Schema(ManagerSchema):
        grid_filename: Path = Field(
            description=(
                "A Geopackage with a collection of bounding boxes and attributes that will define EOPatches. In case "
                "bounding boxes are in multiple CRS then each Geopackage layer should contain bounding boxes from one "
                "CRS."
            ),
        )

    # TODO: The AreaManager needs to be reworked until this is no longer an issue
    config: Schema  # type: ignore[assignment]

    def get_area_dataframe(self, *, crs: CRS = CRS.WGS84, **_: Any) -> gpd.GeoDataFrame:
        """Provides a single dataframe that defines an AOI

        :param crs: A CRS of the dataframe
        """
        grid = self.get_grid()
        return concat_gdf(grid, reproject_crs=crs)

    def get_area_geometry(self, *, crs: CRS = CRS.WGS84, **_: Any) -> Geometry:
        """Provides AOI geometry by joining grid geometries

        :param crs: A CRS in which grid geometries will be joined
        """
        area_filename = self._construct_file_path(prefix="area")

        if self.storage.filesystem.exists(area_filename):
            return self._load_area_geometry(area_filename)

        area_df = self.get_area_dataframe(crs=crs)

        LOGGER.info("Calculating a unary union of the geometries")
        area_shape = shapely.ops.unary_union(area_df.geometry)
        area_geometry = Geometry(area_shape, crs)

        self._save_area_geometry(area_geometry, area_filename)
        return area_geometry

    def get_grid(self, *, add_bbox_column: bool = False, **_: Any) -> List[gpd.GeoDataFrame]:
        """Provides a grid of bounding boxes which divide the AOI

        :param add_bbox_column: If `True` a new column with BBox objects will be added.
        :return: A list of GeoDataFrames containing bounding box geometries and their info. Bounding boxes are divided
            into GeoDataFrames per CRS.
        """
        grid_filename = fs.path.combine(self.storage.get_input_data_folder(), self.config.grid_filename)
        grid = self._load_grid(grid_filename)

        if add_bbox_column:
            self._add_bbox_column(grid)

        return grid

    @staticmethod
    def has_region_filter(*_: Any, **__: Any) -> bool:
        """Doesn't support filtering by a region"""
        return False

    def _construct_file_path(self, *, prefix: str, suffix: str = "gpkg", **_: Any) -> str:
        """Provides a file path of a cached file"""
        input_filename = self.config.grid_filename.rsplit(".", 1)[0]
        input_filename = fs.path.basename(input_filename)

        filename_params = [
            prefix,
            input_filename,
            self.__class__.__name__,
        ]

        filename = "_".join(map(str, filename_params))
        filename = filename.replace(" ", "_")

        return fs.path.combine(self.storage.get_cache_folder(), f"{filename}.{suffix}")



"""
Utilities for exporting data
"""
import logging
from typing import Any, Dict, List, Literal, Optional

import geopandas as gpd
import numpy as np
from fs.base import FS

from sentinelhub import CRS, BBox

from .fs import LocalFile

LOGGER = logging.getLogger(__name__)


def export_grid_stats(
    stats_list: List[Dict[str, object]],
    bbox_list: List[BBox],
    path: str,
    filesystem: Optional[FS] = None,
    names: Optional[List[str]] = None,
    geopandas_backend: Literal["fiona", "pyogrio"] = "fiona",
) -> None:
    """Exports stats per each bounding box (i.e. EOPatch) into a Geopackage file

    :param stats_list: Dictionaries of statistical values and names, one dictionary per each bounding box
    :param bbox_list: Bounding boxes that correspond to stats in `stats_list`
    :param path: A path to a Geopackage file into which data will be exported. In case filesystem object is provided it
        should be a relative path otherwise an absolute path
    :param filesystem: A filesystem object.
    :param names: If provided these names will be included as an additional column in the exported Geopackage
    """
    if len(stats_list) != len(bbox_list):
        raise ValueError(
            f"stats_list and bbox_list should have the same length but found {len(stats_list)} and {len(bbox_list)}"
        )
    if names and len(stats_list) != len(names):
        raise ValueError(
            f"stats_list and names should have the same length but found {len(stats_list)} and {len(names)}"
        )

    data_per_crs: Dict[CRS, List[Dict[str, Any]]] = {}
    for index, (stats, bbox) in enumerate(zip(stats_list, bbox_list)):
        data_dict = {"geometry": bbox.geometry, **stats}
        if names:
            data_dict = {"name": names[index], **data_dict}

        data_per_crs[bbox.crs] = data_per_crs.get(bbox.crs, [])
        data_per_crs[bbox.crs].append(data_dict)

    grid = []
    for crs, data in data_per_crs.items():
        gdf = gpd.GeoDataFrame(data=data, crs=crs.pyproj_crs())

        # Cast objects to strings, otherwise they cannot be serialized into a Geopackage
        for column_name in gdf.columns:
            if gdf[column_name].dtype == np.dtype("object"):
                gdf[column_name] = gdf[column_name].astype(str)

        grid.append(gdf)

    with LocalFile(path, mode="w", filesystem=filesystem) as local_file:
        for gdf in grid:
            gdf.to_file(
                local_file.path,
                driver="GPKG",
                encoding="utf-8",
                layer=f"Stats for EPSG:{gdf.crs.to_epsg()}",
                engine=geopandas_backend,
            )
    LOGGER.info("Saved stats to %s", path)



"""
Module where base Pipeline class is implemented
"""
import datetime as dt
import logging
import time
import uuid
import warnings
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar

from eolearn.core import CreateEOPatchTask, EOExecutor, EONode, EOWorkflow, LoadTask, SaveTask, WorkflowResults
from eolearn.core.extra.ray import RayExecutor

from ..utils.meta import import_object
from ..utils.ray import handle_ray_connection
from ..utils.types import ProcessingType
from .area.base import AreaManager
from .base import EOGrowObject
from .config import RawConfig
from .eopatch import EOPatchManager
from .logging import EOExecutionFilter, EOExecutionHandler, LoggingManager
from .schemas import ManagerSchema, PipelineSchema
from .storage import StorageManager

Self = TypeVar("Self", bound="Pipeline")

LOGGER = logging.getLogger(__name__)


class Pipeline(EOGrowObject):
    """A base class for execution of processing procedures which may or may not include running EOWorkflows, running
    EOExecutions, creating maps, etc.

    The functionalities of this class are:
        - collecting input arguments (either from command line or as an initialization parameter) and parsing them
        - preparing a list of patches
        - preparing execution arguments
        - monitoring the pipeline and reporting
    """

    class Schema(PipelineSchema):
        """Configuration schema, describing input parameters and their types."""

    config: Schema

    def __init__(self, config: Schema, raw_config: Optional[RawConfig] = None):
        """
        :param config: A dictionary with configuration parameters
        :param raw_config: The configuration parameters pre-validation, for logging purposes only
        """
        super().__init__(config)
        self._raw_config = raw_config

        self.pipeline_id = self._new_pipeline_id()
        self.current_execution_name = "<Not executed yet>"

        self.storage: StorageManager = self._load_manager(config.storage)
        self.sh_config = self.storage.sh_config

        self.area_manager: AreaManager = self._load_manager(config.area, storage=self.storage)
        self.eopatch_manager: EOPatchManager = self._load_manager(config.eopatch, area_manager=self.area_manager)
        self.logging_manager: LoggingManager = self._load_manager(config.logging, storage=self.storage)

        self._patch_list: Optional[List[str]] = None

    @classmethod
    def from_raw_config(cls: Type[Self], config: RawConfig, *args: Any, **kwargs: Any) -> Self:
        """Creates an object from a dictionary by constructing a validated config and use it to create the object."""
        validated_config = cls.Schema.parse_obj(config)
        if "raw_config" not in kwargs:
            kwargs["raw_config"] = config
        return cls(validated_config, *args, **kwargs)

    @staticmethod
    def _new_pipeline_id() -> str:
        """Provides a new random uuid of a pipeline"""
        return uuid.uuid4().hex[:10]

    @staticmethod
    def _load_manager(manager_config: ManagerSchema, **manager_params: Any) -> Any:
        """Loads a manager class and back-propagates parsed config

        :param manager_key: A config key name of a sub-config with manager parameters
        :param manager_params: Other parameters to initialize a manager class
        """
        if manager_config.manager is None:
            raise ValueError("Unable to load manager, field `manager` specifying it's class is missing.")
        manager_class = import_object(manager_config.manager)
        manager = manager_class(manager_config, **manager_params)
        return manager

    def get_pipeline_execution_name(self, pipeline_timestamp: str) -> str:
        """Returns the full name of the pipeline execution"""
        return f"{pipeline_timestamp}-{self.__class__.__name__}-{self.pipeline_id}"

    @property
    def patch_list(self) -> List[str]:
        """A property that provides a list of EOPatch names that will be used by the pipeline. The list is calculated
        lazily the first time this property is called.
        """
        if self._patch_list is None:
            self._patch_list = self._prepare_patch_list()
        return self._patch_list

    def _prepare_patch_list(self) -> List[str]:
        """Method which at the initialization prepares the list of EOPatches which will be used"""
        if self.config.input_patch_file is None:
            patch_list = self.eopatch_manager.get_eopatch_filenames(id_list=self.config.patch_list)
        else:
            if self.config.patch_list:
                warnings.warn(
                    "'patch_list' and 'input_patch_file' parameters have both been given, therefore patches "
                    "from input_patch_file will be filtered according to patch_list indices",
                    RuntimeWarning,
                )
            patch_list = self.eopatch_manager.load_eopatch_filenames(
                self.config.input_patch_file, id_list=self.config.patch_list
            )

        if self.config.skip_existing:
            LOGGER.info("Checking which EOPatches can be skipped")
            filtered_patch_list = self.filter_patch_list(patch_list)

            skip_message = (
                "Skipped some EOPatches" if len(filtered_patch_list) < len(patch_list) else "No EOPatches were skipped"
            )
            LOGGER.info("%s, %d / %d remaining", skip_message, len(filtered_patch_list), len(patch_list))

            patch_list = filtered_patch_list

        return patch_list

    def filter_patch_list(self, patch_list: List[str]) -> List[str]:
        """Overwrite this method to specify which EOPatches should be filtered with `skip_existing`"""
        raise NotImplementedError("Method `filter_patch_list` must be implemented in order to use `skip_existing`")

    def get_execution_arguments(self, workflow: EOWorkflow) -> List[Dict[EONode, Dict[str, object]]]:
        """Prepares execution arguments for each eopatch from a list of patches

        :param workflow: A workflow for which arguments will be prepared
        """
        bbox_list = self.eopatch_manager.get_bboxes(eopatch_list=self.patch_list)

        exec_args = []
        nodes = workflow.get_nodes()
        for name, bbox in zip(self.patch_list, bbox_list):
            single_exec_dict: Dict[EONode, Dict[str, Any]] = {}

            for node in nodes:
                if isinstance(node.task, (SaveTask, LoadTask)):
                    single_exec_dict[node] = dict(eopatch_folder=name)

                if isinstance(node.task, CreateEOPatchTask):
                    single_exec_dict[node] = dict(bbox=bbox)

            exec_args.append(single_exec_dict)
        return exec_args

    def _init_processing(self) -> ProcessingType:
        """Figures out which execution mode is used and configures connection to Ray if required."""
        is_connected = handle_ray_connection(self.config.use_ray)
        if is_connected:
            return ProcessingType.RAY
        return ProcessingType.MULTI if self.config.workers > 1 else ProcessingType.SINGLE

    def run_execution(
        self,
        workflow: EOWorkflow,
        exec_args: List[dict],
        eopatch_list: Optional[List[str]] = None,
        **executor_run_params: Any,
    ) -> Tuple[List[str], List[str], List[WorkflowResults]]:
        """A method which runs EOExecutor on given workflow with given execution parameters

        :param workflow: A workflow to be executed
        :param exec_args: A list of dictionaries holding execution arguments
        :param eopatch_list: A custom list of EOPatch names on which execution will run. If not specified, the default
            self.patch_list will be used
        :return: Lists of successfully/unsuccessfully executed EOPatch names and the result of the EOWorkflow execution
        """
        if eopatch_list is None:
            eopatch_list = self.patch_list

        executor_class: Type[EOExecutor]

        execution_kind = self._init_processing()
        if execution_kind is ProcessingType.RAY:
            executor_class = RayExecutor
        else:
            executor_class = EOExecutor
            executor_run_params["workers"] = self.config.workers

        LOGGER.info("Starting %s for %d EOPatches", executor_class.__name__, len(exec_args))
        executor = executor_class(
            workflow,
            exec_args,
            execution_names=eopatch_list,
            save_logs=self.logging_manager.config.save_logs,
            logs_folder=self.logging_manager.get_pipeline_logs_folder(self.current_execution_name),
            filesystem=self.storage.filesystem,
            logs_filter=EOExecutionFilter(ignore_packages=self.logging_manager.config.eoexecution_ignore_packages),
            logs_handler_factory=EOExecutionHandler,
        )
        execution_results = executor.run(**executor_run_params)

        successful_eopatches = [eopatch_list[idx] for idx in executor.get_successful_executions()]
        failed_eopatches = [eopatch_list[idx] for idx in executor.get_failed_executions()]
        LOGGER.info(
            "%s finished with %d / %d success rate",
            executor_class.__name__,
            len(successful_eopatches),
            len(successful_eopatches) + len(failed_eopatches),
        )

        if self.logging_manager.config.save_logs:
            executor.make_report(include_logs=self.logging_manager.config.include_logs_to_report)
            LOGGER.info("Saved EOExecution report to %s", executor.get_report_path(full_path=True))

        return successful_eopatches, failed_eopatches, execution_results

    def run(self) -> None:
        """Call this method to run any pipeline"""
        timestamp = dt.datetime.utcnow().strftime("%Y-%m-%dT%H-%M-%SZ")
        self.current_execution_name = self.get_pipeline_execution_name(timestamp)

        handlers = self.logging_manager.start_logging(self.current_execution_name)
        try:
            self.logging_manager.update_pipeline_report(
                pipeline_execution_name=self.current_execution_name,
                pipeline_config=self.config,
                pipeline_raw_config=self._raw_config,
                pipeline_id=self.pipeline_id,
                pipeline_timestamp=timestamp,
            )

            LOGGER.info("Running %s", self.__class__.__name__)

            pipeline_start = time.time()
            finished, failed = self.run_procedure()
            elapsed_time = time.time() - pipeline_start

            if failed:
                LOGGER.info(
                    "Pipeline finished with some errors! Check %s",
                    self.logging_manager.get_pipeline_logs_folder(self.current_execution_name, full_path=True),
                )
            else:
                LOGGER.info("Pipeline finished successfully!")

            self.logging_manager.update_pipeline_report(
                pipeline_execution_name=self.current_execution_name,
                pipeline_raw_config=self._raw_config,
                pipeline_config=self.config,
                pipeline_id=self.pipeline_id,
                pipeline_timestamp=timestamp,
                elapsed_time=elapsed_time,
            )

            finished = self.eopatch_manager.parse_eopatch_list(finished)
            failed = self.eopatch_manager.parse_eopatch_list(failed)
            self.logging_manager.save_eopatch_execution_status(
                pipeline_execution_name=self.current_execution_name, finished=finished, failed=failed
            )
        finally:
            self.logging_manager.stop_logging(handlers)

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Execution procedure of pipeline. Can be overridden if needed.

        By default, builds the workflow by using a `build_workflow` method, which must be additionally implemented.

        :return: A list of successfully executed EOPatch names and a list of unsuccessfully executed EOPatch names
        """
        if not hasattr(self, "build_workflow"):
            raise NotImplementedError(
                "Default implementation of run_procedure method requires implementation of build_workflow method"
            )
        workflow = self.build_workflow()
        exec_args = self.get_execution_arguments(workflow)

        finished, failed, _ = self.run_execution(workflow, exec_args)
        return finished, failed
